/*
 * Built-in function handling routines.
 * This only contains the table of built-in functions and the support
 * routines for calling them.  The functions themselves are in other files.
 *
 * Copyright (C) 2009  David I. Bell
 *
 * Calc is open software; you can redistribute it and/or modify it under
 * the terms of the version 2.1 of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Calc is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU Lesser General
 * Public License for more details.
 *
 * A copy of version 2.1 of the GNU Lesser General Public License is
 * distributed with calc under the filename COPYING-LGPL.  You should have
 * received a copy with calc; if not, write to Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 */

#include "calc.h"
#include "nodeinfo.h"
#include "builtinfunc.h"
#include "token.h"


/*
 * List of primitive built-in functions
 *
 * NOTE:  This table is also used by the help/Makefile builtin rule to
 *	  form the builtin help file.  This rule will cause a sed script
 *	  to strip this table down into a just the information needed
 *	  to print builtin function list: name, minArgs, maxArgs
 *	  and desc.  All other struct elements will be converted to 0.
 *	  The sed script expects to find entries of the form:
 *
 *		{"...", number, number, stuff, stuff, stuff, stuff,
 *		 "...."},
 *
 *	  please keep this table in that form.
 *
 *	  For nice output, when the description of function (desc)
 *	  gets too long (extends into col 79) you should chop the
 *	  line and add "\n\t\t    ", thats newline, 2 tabs a 4 spaces.
 *	  For example the description:
 *
 *		... very long description that goes beyond col 79
 *
 *	  should be written as:
 *
 *		"... very long description that\n\t\t    goes beyond col 79"},
 *
 * fields:
 *	name		name of built-in function
 *	minArgs		minimum number of arguments
 *	maxArgs		maximum number of arguments
 *	flags		special handling flags
 *	opcode		opcode which makes the call quick
 *	numberFunc	routine to calculate numeric function
 *	valueFunc	routine to calculate general values
 *	desc		description of function
 */
static const BuiltinFuncInfo builtins[] =
{
	{"abs", 1, 2, 0, OP_ABSOLUTE_VALUE, 0, 0,
	 "absolute value within accuracy b"},
	{"access", 1, 2, 0, OP_NONE, 0, f_access,
	 "determine accessibility of file a for mode b"},
	{"acos", 1, 2, FE, OP_NONE, 0, f_acos,
	 "arccosine of a within accuracy b"},
	{"acosh", 1, 2, FE, OP_NONE, 0, f_acosh,
	 "inverse hyperbolic cosine of a within accuracy b"},
	{"acot", 1, 2, FE, OP_NONE, 0, f_acot,
	 "arccotangent of a within accuracy b"},
	{"acoth", 1, 2, FE, OP_NONE, 0, f_acoth,
	 "inverse hyperbolic cotangent of a within accuracy b"},
	{"acsc", 1, 2, FE, OP_NONE, 0, f_acsc,
	 "arccosecant of a within accuracy b"},
	{"acsch", 1, 2, FE, OP_NONE, 0, f_acsch,
	 "inverse csch of a within accuracy b"},
	{"agd", 1, 2, 0, OP_NONE, 0, f_agd,
	 "inverse gudermannian function"},	 
	{"append", 1, IN, FA, OP_NONE, 0, f_listappend,
	 "append values to end of list"},
	{"appr", 1, 3, 0, OP_NONE, 0, f_appr,
	 "approximate a by multiple of b using rounding c"},
	{"arg", 1, 2, 0, OP_NONE, 0, f_arg,
	 "argument (the angle) of complex number"},
	{"argv", 0, 1, 0, OP_NONE, 0, f_argv,
	 "calc argc or argv string"},
	{"asec", 1, 2, FE, OP_NONE, 0, f_asec,
	 "arcsecant of a within accuracy b"},
	{"asech", 1, 2, FE, OP_NONE, 0, f_asech,
	 "inverse hyperbolic secant of a within accuracy b"},
	{"asin", 1, 2, FE, OP_NONE, 0, f_asin,
	 "arcsine of a within accuracy b"},
	{"asinh", 1, 2, FE, OP_NONE, 0, f_asinh,
	 "inverse hyperbolic sine of a within accuracy b"},
	{"assoc", 0, 0, 0, OP_NONE, 0, f_assoc,
	 "create new association array"},
	{"atan", 1, 2, FE, OP_NONE, 0, f_atan,
	 "arctangent of a within accuracy b"},
	{"atan2", 2, 3, FE, OP_NONE, qatan2, 0,
	 "angle to point (b,a) within accuracy c"},
	{"atanh", 1, 2, FE, OP_NONE, 0, f_atanh,
	 "inverse hyperbolic tangent of a within accuracy b"},
	{"avg", 0, IN, 0, OP_NONE, 0, f_avg,
	 "arithmetic mean of values"},
	{"base", 0, 1, 0, OP_NONE, f_base, 0,
	 "set default output base"},
	{"base2", 0, 1, 0, OP_NONE, f_base2, 0,
	 "set default secondary output base"},
	{"bernoulli", 1, 1, 0, OP_NONE, 0, f_bern,
	 "Bernoulli number for index a"},
	{"bit", 2, 2, 0, OP_BIT, 0, 0,
	 "whether bit b in value a is set"},
	{"blk", 0, 3, 0, OP_NONE, 0, f_blk,
	 "block with or without name, octet number, chunksize"},
	{"blkcpy", 2, 5, 0, OP_NONE, 0, f_blkcpy,
	 "copy value to/from a block: blkcpy(d,s,len,di,si)"},
	{"blkfree", 1, 1, 0, OP_NONE, 0, f_blkfree,
	 "free all storage from a named block"},
	{"blocks", 0, 1, 0, OP_NONE, 0, f_blocks,
	 "named block with specified index, or null value"},
	{"bround", 1, 3, 0, OP_NONE, 0, f_bround,
	 "round value a to b number of binary places"},
	{"btrunc", 1, 2, 0, OP_NONE, f_btrunc, 0,
	 "truncate a to b number of binary places"},
	{"calclevel", 0, 0, 0, OP_NONE, f_calclevel, 0,
	 "current calculation level"},
	{"calc_tty", 0, 0, 0, OP_NONE, 0, f_calc_tty,
	 "set tty for interactivity"},
	{"callername", 1, 1, 0, OP_NONE, 0, f_callername,
	 "caller function name"},
	{"callerline", 1, 1, 0, OP_NONE, 0, f_callerline,
	 "caller line number"},
	{"catalan", 1, 1, 0, OP_NONE, 0, f_catalan,
	 "catalan number for index a"},
	{"ceil", 1, 1, 0, OP_NONE, 0, f_ceil,
	 "smallest integer greater than or equal to number"},
	{"cfappr", 1, 3, 0, OP_NONE, f_cfappr, 0,
	 "approximate a within accuracy b using\n\t\t    continued fractions"},
	{"cfsim", 1, 2, 0, OP_NONE, f_cfsim, 0,
	 "simplify number using continued fractions"},
	{"char", 1, 1, 0, OP_NONE, 0, f_char,
	 "character corresponding to integer value"},
	{"cmdbuf", 0, 0, 0, OP_NONE, 0, f_cmdbuf,
	 "command buffer"},
	{"cmp", 2, 2, 0, OP_COMPARE, 0, 0,
	 "compare values returning -1, 0, or 1"},
	{"comb", 2, 2, 0, OP_NONE, 0, f_comb,
	 "combinatorial number a!/b!(a-b)!"},
	{"config", 1, 2, 0, OP_CONFIG, 0, 0,
	 "set or read configuration value"},
	{"conj", 1, 1, 0, OP_CONJUGATE, 0, 0,
	 "complex conjugate of value"},
	{"copy", 2, 5, 0, OP_NONE, 0, f_copy,
	 "copy value to/from a block: copy(s,d,len,si,di)"},
	{"cos", 1, 2, 0, OP_NONE, 0, f_cos,
	 "cosine of value a within accuracy b"},
	{"cosh", 1, 2, FE, OP_NONE, 0, f_cosh,
	 "hyperbolic cosine of a within accuracy b"},
	{"cot", 1, 2, FE, OP_NONE, 0, f_cot,
	 "cotangent of a within accuracy b"},
	{"coth", 1, 2, FE, OP_NONE, 0, f_coth,
	 "hyperbolic cotangent of a within accuracy b"},
	{"count", 2, 2, 0, OP_NONE, 0, f_count,
	 "count list/matrix elements satisfying some condition"},
	{"cp", 2, 2, 0, OP_NONE, 0, f_cp,
	 "cross product of two vectors"},
	{"csc", 1, 2, FE, OP_NONE, qcsc, 0,
	 "cosecant of a within accuracy b"},
	{"csch", 1, 2, FE, OP_NONE, 0, f_csch,
	 "hyperbolic cosecant of a within accuracy b"},
	{"ctime", 0, 0, 0, OP_NONE, 0, f_ctime,
	 "date and time as string"},
	{"custom", 0, IN, 0, OP_NONE, 0, f_custom,
	 "custom builtin function interface"},
	{"delete", 2, 2, FA, OP_NONE, 0, f_listdelete,
	 "delete element from list a at position b"},
	{"den", 1, 1, 0, OP_DENOMINATOR, qden, 0,
	 "denominator of fraction"},
	{"det", 1, 1, 0, OP_NONE, 0, f_det,
	 "determinant of matrix"},
	{"digit", 2, 3, 0, OP_NONE, 0, f_digit,
	 "digit at specified decimal place of number"},
	{"digits", 1, 2, 0, OP_NONE, 0, f_digits,
	 "number of digits in number"},
	{"display", 0, 1, 0, OP_NONE, f_display, 0,
	 "number of decimal digits for displaying numbers"},	 
	{"dp", 2, 2, 0, OP_NONE, 0, f_dp,
	 "dot product of two vectors"},
	{"epsilon", 0, 1, 0, OP_EPSILON, 0, 0,
	 "set or read allowed error for real calculations"},
	{"errcount", 0, 1, 0, OP_NONE, 0, f_errcount,
	 "set or read error count"},
	{"errmax", 0, 1, 0, OP_NONE, 0, f_errmax,
	 "set or read maximum for error count"},
	{"errno", 0, 1, 0, OP_NONE, 0, f_errno,
	 "system error message"},
	{"error", 0, 6, 0, OP_NONE, 0, f_error,
	 "generate or return error value"},
	{"euler", 1, 1, 0, OP_NONE, 0, f_euler,
	 "Euler number"},
	{"eval", 1, 1, 0, OP_NONE, 0, f_eval,
	 "evaluate expression from string to value"},
	{"exp", 1, 2, 0, OP_NONE, 0, f_exp,
	 "exponential of value a within accuracy b"},
	{"expr", 1, 2, 0, OP_NONE, 0, f_expr,
	 "create symbolic expression value"},
	{"factor", 1, 3, 0, OP_NONE, f_factor, 0,
	 "lowest prime factor < b of a, return c if error"},
	{"fcnt", 2, 2, 0, OP_NONE, f_faccnt, 0,
	 "count of times one number divides another"},
	{"fib", 1, 1, 0, OP_NONE, qfib, 0,
	 "Fibonacci number F(n)"},
	{"forall", 2, 2, 0, OP_NONE, 0, f_forall,
	 "do function for all elements of list or matrix"},
	{"frem", 2, 2, 0, OP_NONE, qfacrem, 0,
	 "number with all occurrences of factor removed"},
	{"fact", 1, 1, 0, OP_FACTORIAL, 0, 0,
	 "factorial"},
	{"fclose", 0, IN, 0, OP_NONE, 0, f_fclose,
	 "close file"},
	{"feof", 1, 1, FO, OP_NONE, 0, f_feof,
	 "whether EOF reached for file"},
	{"ferror", 1, 1, FO, OP_NONE, 0, f_ferror,
	 "whether error occurred for file"},
	{"fflush", 0, IN, FO, OP_NONE, 0, f_fflush,
	 "flush output to file(s)"},
	{"fgetc", 1, 1, FO, OP_NONE, 0, f_fgetc,
	 "read next char from file"},
	{"fgetfield", 1, 1, FO, OP_NONE, 0, f_fgetfield,
	 "read next white-space delimited field from file"},
	{"fgetline", 1, 1, FO, OP_NONE, 0, f_fgetline,
	 "read next line from file, newline removed"},
	{"fgets", 1, 1, FO, OP_NONE, 0, f_fgets,
	 "read next line from file, newline is kept"},
	{"fgetstr", 1, 1, FO, OP_NONE, 0, f_fgetstr,
	 "read next null-terminated string from file, null character is kept"},
	{"files", 0, 1, 0, OP_NONE, 0, f_files,
	 "return opened file or max number of opened files"},
	{"floor", 1, 1, 0, OP_NONE, 0, f_floor,
	 "greatest integer less than or equal to number"},
	{"fopen", 2, 2, 0, OP_NONE, 0, f_fopen,
	 "open file name a in mode b"},
	{"fprintf", 2, IN, FO, OP_NONE, 0, f_fprintf,
	 "print formatted output to opened file"},
	{"fputc", 2, 2, FO, OP_NONE, 0, f_fputc,
	 "write a character to a file"},
	{"fputs", 2, IN, FO, OP_NONE, 0, f_fputs,
	 "write one or more strings to a file"},
	{"fputstr", 2, IN, FO, OP_NONE, 0, f_fputstr,
	 "write one or more null-terminated strings to a file"},
	{"freopen", 2, 3, FO, OP_NONE, 0, f_freopen,
	 "reopen a file stream to a named file"},
	{"fscan", 2, IN, FA, OP_NONE, 0, f_fscan,
	 "scan a file for assignments to one or more variables"},
	{"fscanf", 2, IN, FA, OP_NONE, 0, f_fscanf,
	 "formatted scan of a file for assignment to one or more variables"},
	{"fseek", 2, 3, FO, OP_NONE, 0, f_fseek,
	 "seek to position b (offset from c) in file a"},
	{"fsize", 1, 1, FO, OP_NONE, 0, f_fsize,
	 "return the size of the file"},
	{"ftell", 1, 1, FO, OP_NONE, 0, f_ftell,
	 "return the file position"},
	{"frac", 1, 1, 0, OP_FRACTIONAL_PART, qfrac, 0,
	 "fractional part of value"},
	{"free", 0, IN, FA, OP_NONE, 0, f_free,
	 "free listed or all global variables"},
	{"freebernoulli", 0, 0, 0, OP_NONE, 0, f_freebern,
	 "free stored Benoulli numbers"},
	{"freeeuler", 0, 0, 0, OP_NONE, 0, f_freeeuler,
	 "free stored Euler numbers"},
	{"freeglobals", 0, 0, 0, OP_NONE, 0, f_freeglobals,
	 "free all global and visible static variables"},
	{"freeredc", 0, 0, 0, OP_NONE, 0, f_freeredc,
	 "free redc data cache"},
	{"freestatics", 0, 0, 0, OP_NONE, 0, f_freestatics,
	 "free all unscoped static variables"},
	{"func", 0, 1, 0, OP_NONE, 0, f_func,
	 "return reference to a named (or current) function or method"},
	{"gcd", 1, IN, 0, OP_NONE, f_gcd, 0,
	 "greatest common divisor"},
	{"gcdrem", 2, 2, 0, OP_NONE, qgcdrem, 0,
	 "a divided repeatedly by gcd with b"},
	{"gd", 1, 2, 0, OP_NONE, 0, f_gd,
	 "gudermannian function"},
	{"getenv", 1, 1, 0, OP_NONE, 0, f_getenv,
	 "value of environment variable (or NULL)"},
	{"hash", 1, IN, 0, OP_NONE, 0, f_hash,
	 "return non-negative hash value for one or\n\t\t    more values"},
	{"head", 2, 2, 0, OP_NONE, 0, f_head,
	 "return list of specified number at head of a list"},
	{"highbit", 1, 1, 0, OP_HIGH_BIT, 0, 0,
	 "high bit number in base 2 representation"},
	{"hmean", 0, IN, 0, OP_NONE, 0, f_hmean,
	 "harmonic mean of values"},
	{"hnrmod", 4, 4, 0, OP_NONE, f_hnrmod, 0,
	 "v mod h*2^n+r, h>0, n>0, r = -1, 0 or 1"},
	{"hypot", 2, 3, FE, OP_NONE, qhypot, 0,
	 "hypotenuse of right triangle within accuracy c"},
	{"ilog", 2, 2, 0, OP_NONE, 0, f_ilog,
	 "integral log of one number with another"},
	{"ilog10", 1, 1, 0, OP_NONE, 0, f_ilog10,
	 "integral log of a number base 10"},
	{"ilog2", 1, 1, 0, OP_NONE, 0, f_ilog2,
	 "integral log of a number base 2"},
	{"im", 1, 1, 0, OP_IMAGINARY_PART, 0, 0,
	 "imaginary part of complex number"},
	{"indices", 2, 2, 0, OP_NONE, 0, f_indices,
	 "indices of a specified assoc or mat value"},
	{"inputlevel", 0, 0, 0, OP_NONE, f_inputlevel, 0,
	 "current input depth"},
	{"insert", 2, IN, FA, OP_NONE, 0, f_listinsert,
	 "insert values c ... into list a at position b"},
	{"int", 1, 1, 0, OP_INTEGER_PART, qint, 0,
	 "integer part of value"},
	{"inverse", 1, 1, 0, OP_INVERT, 0, 0,
	 "multiplicative inverse of value"},
	{"iroot", 2, 2, 0, OP_NONE, qiroot, 0,
	 "integer b'th root of a"},
	{"isassoc", 1, 1, 0, OP_IS_ASSOCIATION, 0, 0,
	 "whether a value is an association"},
	{"isatty", 1, 1, 0, OP_NONE, 0, f_isatty,
	 "whether a file is a tty"},
	{"isblk", 1, 1, 0, OP_IS_BLOCK, 0, 0,
	 "whether a value is a memory block"},
	{"isconfig", 1, 1, 0, OP_IS_CONFIG, 0, 0,
	 "whether a value is a config state"},
	{"isdefined", 1, 1, 0, OP_NONE, 0, f_isDefined,
	 "whether a string names a function"},
	{"iserror", 1, 1, 0, OP_IS_ERROR, 0, 0,
	 "whether a value is an error"},
	{"iseven", 1, 1, 0, OP_IS_EVEN, 0, 0,
	 "whether a value is an even integer"},
	{"isfile", 1, 1, 0, OP_IS_FILE, 0, 0,
	 "whether a value is a file"},
	{"isfunc", 1, 1, 0, OP_IS_FUNCTION, 0, 0,
	 "whether a value is a function or method reference"},
	{"ishash", 1, 1, 0, OP_IS_HASH, 0, 0,
	 "whether a value is a hash state"},
	{"isident", 1, 1, 0, OP_NONE, 0, f_isident,
	 "returns 1 if identity matrix"},
	{"isint", 1, 1, 0, OP_IS_INTEGER, 0, 0,
	 "whether a value is an integer"},
	{"islist", 1, 1, 0, OP_IS_LIST, 0, 0,
	 "whether a value is a list"},
	{"ismat", 1, 1, 0, OP_IS_MATRIX, 0, 0,
	 "whether a value is a matrix"},
	{"ismult", 2, 2, 0, OP_NONE, f_ismult, 0,
	 "whether a is a multiple of b"},
	{"isnull", 1, 1, 0, OP_IS_NULL, 0, 0,
	 "whether a value is the null value"},
	{"isnum", 1, 1, 0, OP_IS_NUMBER, 0, 0,
	 "whether a value is a number"},
	{"isobj", 1, 1, 0, OP_IS_OBJECT, 0, 0,
	 "whether a value is an object"},
	{"isobjtype", 1, 1, 0, OP_IS_OBJECT_TYPE, 0,0,
	 "whether a string names an object type"},
	{"isodd", 1, 1, 0, OP_IS_ODD, 0, 0,
	 "whether a value is an odd integer"},
	{"isprime", 1, 2, 0, OP_NONE, f_isprime, 0,
	 "whether a is a small prime, return b if error"},
	{"isqrt", 1, 1, 0, OP_NONE, qisqrt, 0,
	 "integer part of square root"},
	{"isrand", 1, 1, 0, OP_IS_RAND, 0, 0,
	 "whether a value is a additive 55 state"},
	{"israndom", 1, 1, 0, OP_IS_RANDOM, 0, 0,
	 "whether a value is a Blum state"},
	{"isreal", 1, 1, 0, OP_IS_REAL, 0, 0,
	 "whether a value is a real number"},
	{"isrel", 2, 2, 0, OP_NONE, f_isrel, 0,
	 "whether two numbers are relatively prime"},
	{"isstr", 1, 1, 0, OP_IS_STRING, 0, 0,
	 "whether a value is a string"},
	{"issimple", 1, 1, 0, OP_IS_SIMPLE, 0, 0,
	 "whether value is a simple type"},
	{"issq", 1, 1, 0, OP_NONE, f_issquare, 0,
	 "whether or not number is a square"},
	{"istype", 2, 2, 0, OP_IS_TYPE, 0, 0,
	 "whether the type of a is same as the type of b"},
	{"jacobi", 2, 2, 0, OP_NONE, qjacobi, 0,
	 "-1 => a is not quadratic residue mod b\n\t\t    1 => b is composite, or a is quad residue of b"},
	{"join", 1, IN, 0, OP_NONE, 0, f_join,
	 "join one or more lists into one list"},
	{"lcm", 1, IN, 0, OP_NONE, f_lcm, 0,
	 "least common multiple"},
	{"lcmfact", 1, 1, 0, OP_NONE, qlcmfact, 0,
	 "lcm of all integers up till number"},
	{"lfactor", 2, 2, 0, OP_NONE, qlowfactor, 0,
	 "lowest prime factor of a in first b primes"},
	{"links", 1, 1, 0, OP_NONE, 0, f_links,
	 "return link count for a value"},
	{"list", 0, IN, 0, OP_NONE, 0, f_list,
	 "create list of specified values"},
	{"ln", 1, 2, 0, OP_NONE, 0, f_ln,
	 "natural logarithm of value a within accuracy b"},
	{"lowbit", 1, 1, 0, OP_LOW_BIT, 0, 0,
	 "low bit number in base 2 representation"},
	{"ltol", 1, 2, FE, OP_NONE, f_legtoleg, 0,
	 "leg-to-leg of unit right triangle (sqrt(1 - a^2))"},
	{"makelist", 1, 1, 0, OP_NONE, 0, f_makelist,
	 "create a list with a null elements"},
	{"matdim", 1, 1, 0, OP_NONE, 0, f_matdim,
	 "number of dimensions of matrix"},
	{"matfill", 2, 3, FA, OP_NONE, 0, f_matfill,
	 "fill matrix with value b (value c on diagonal)"},
	{"matmax", 2, 2, 0, OP_NONE, 0, f_matmax,
	 "maximum index of matrix a dim b"},
	{"matmin", 2, 2, 0, OP_NONE, 0, f_matmin,
	 "minimum index of matrix a dim b"},
	{"matsum", 1, 1, 0, OP_NONE, 0, f_matsum,
	 "sum the numeric values in a matrix"},
	{"mattrace", 1, 1, 0, OP_NONE, 0, f_mattrace,
	 "return the trace of a square matrix"},
	{"mattrans", 1, 1, 0, OP_NONE, 0, f_mattrans,
	 "transpose of matrix"},
	{"max", 0, IN, 0, OP_NONE, 0, f_max,
	 "maximum value"},
	{"md5", 0, IN, 0, OP_NONE, 0, f_md5,
	 "MD5 Hash Algorithm"},
	{"memsize", 1, 1, 0, OP_NONE, 0, f_memsize,
	 "number of octets used by the value, including overhead"},
	{"meq", 3, 3, 0, OP_NONE, f_meq, 0,
	 "whether a and b are equal modulo c"},
	{"min", 0, IN, 0, OP_NONE, 0, f_min,
	 "minimum value"},
	{"minv", 2, 2, 0, OP_NONE, qminv, 0,
	 "inverse of a modulo b"},
	{"mmin", 2, 2, 0, OP_NONE, 0, f_mmin,
	 "a mod b value with smallest abs value"},
	{"mne", 3, 3, 0, OP_NONE, f_mne, 0,
	 "whether a and b are not equal modulo c"},
	{"mod", 2, 3, 0, OP_NONE, 0, f_mod,
	 "residue of a modulo b, rounding type c"},
	{"modify", 2, 2, FA, OP_NONE, 0, f_modify,
	 "modify elements of a list or matrix"},
	{"name", 1, 1, 0, OP_NONE, 0, f_name,
	 "name assigned to block, file, or function"},
	{"near", 2, 3, 0, OP_NONE, f_near, 0,
	 "sign of (abs(a-b) - c)"},
	{"newerror", 0, 1, 0, OP_NONE, 0, f_newerror,
	 "create new error type with message a"},
	{"nextcand", 1, 5, 0, OP_NONE, f_nextcand, 0,
	 "smallest value == d mod e > a, ptest(a,b,c) true"},
	{"nextprime", 1, 2, 0, OP_NONE, f_nprime, 0,
	 "return next small prime, return b if err"},
	{"norm", 1, 1, 0, OP_NORM, 0, 0,
	 "norm of a value (square of absolute value)"},
	{"null", 0, 1, 0, OP_SET_NULL, 0, 0,
	 "the null value or set a value to null"},
	{"num", 1, 1, 0, OP_NUMERATOR, qnum, 0,
	 "numerator of fraction"},
	{"ord", 1, 1, 0, OP_NONE, 0, f_ord,
	 "integer corresponding to character value"},
	{"param", 1, 1, 0, OP_ARG_VALUE, 0, 0,
	 "value of parameter n (or parameter count if n\n\t\t    is zero)"},
	{"perm", 2, 2, 0, OP_NONE, qperm, 0,
	 "permutation number a!/(a-b)!"},
	{"prevcand", 1, 5, 0, OP_NONE, f_prevcand, 0,
	 "largest value == d mod e < a, ptest(a,b,c) true"},
	{"prevprime", 1, 2, 0, OP_NONE, f_pprime, 0,
	 "return previous small prime, return b if err"},
	{"pfact", 1, 1, 0, OP_NONE, qpfact, 0,
	 "product of primes up till number"},
	{"pi", 0, 1, FE, OP_NONE, qpi, 0,
	 "value of pi accurate to within epsilon"},
	{"pix", 1, 2, 0, OP_NONE, f_pix, 0,
	 "number of primes <= a < 2^32, return b if error"},
	{"places", 1, 2, 0, OP_NONE, 0, f_places,
	 "places after decimal point (-1 if infinite)"},
	{"pmod", 3, 3, 0, OP_NONE, qpowermod,0,
	 "mod of a power (a ^ b (mod c))"},
	{"polar", 2, 3, 0, OP_NONE, 0, f_polar,
	 "complex value of polar coordinate (a * exp(b*1i))"},
	{"poly", 1, IN, 0, OP_NONE, 0, f_poly,
	 "evaluates a polynomial given its coefficients or coefficient-list"},
	{"pop", 1, 1, FA, OP_NONE, 0, f_listpop,
	 "pop value from front of list"},
	{"popcnt", 1, 2, 0, OP_NONE, f_popcnt, 0,
	 "number of bits in a that match b (or 1)"},
	{"power", 2, 3, 0, OP_NONE, 0, f_power,
	 "value a raised to the power b within accuracy c"},
	{"ptest", 1, 3, 0, OP_NONE, f_primetest, 0,
	 "probabilistic primality test"},
	{"printf", 1, IN, 0, OP_NONE, 0, f_printf,
	 "print formatted output to stdout"},
	{"prompt", 1, 1, 0, OP_NONE, 0, f_prompt,
	 "prompt for input line using value a"},
	{"properties", 0, 2, 0, OP_NONE, 0, f_properties,
	 "get or set global propery value"},
	{"push", 1, IN, FA, OP_NONE, 0, f_listpush,
	 "push values onto front of list"},
	{"putenv", 1, 2, 0, OP_NONE, 0, f_putenv,
	 "define an environment variable"},
	{"quo", 2, 3, 0, OP_NONE, 0, f_quo,
	 "integer quotient of a by b, rounding type c"},
	{"quomod", 4, 4, 0, OP_QUOTIENT_MOD, 0, 0,
	 "set c and d to quotient and remainder of a\n\t\t    divided by b"},
	{"rand", 0, 2, 0, OP_NONE, f_rand, 0,
	 "additive 55 random number [0,2^64), [0,a), or [a,b)"},
	{"randbit", 0, 1, 0, OP_NONE, f_randbit, 0,
	 "additive 55 random number [0,2^a)"},
	{"random", 0, 2, 0, OP_NONE, f_random, 0,
	 "Blum-Blum-Shub random number [0,2^64), [0,a), or [a,b)"},
	{"randombit", 0, 1, 0, OP_NONE, f_randombit, 0,
	 "Blum-Blum-Sub random number [0,2^a)"},
	{"randperm", 1, 1, 0, OP_NONE, 0, f_randperm,
	 "random permutation of a list or matrix"},
	{"rcin", 2, 2, 0, OP_NONE, qredcin, 0,
	 "convert normal number a to REDC number mod b"},
	{"rcmul", 3, 3, 0, OP_NONE, qredcmul, 0,
	 "multiply REDC numbers a and b mod c"},
	{"rcout", 2, 2, 0, OP_NONE, qredcout, 0,
	 "convert REDC number a mod b to normal number"},
	{"rcpow", 3, 3, 0, OP_NONE, qredcpower, 0,
	 "raise REDC number a to power b mod c"},
	{"rcsq", 2, 2, 0, OP_NONE, qredcsquare, 0,
	 "square REDC number a mod b"},
	{"re", 1, 1, 0, OP_REAL_PART, 0, 0,
	 "real part of complex number"},
	{"remove", 1, 1, FA, OP_NONE, 0, f_listremove,
	 "remove value from end of list"},
	{"reverse", 1, 1, 0, OP_NONE, 0, f_reverse,
	 "reverse a copy of a matrix or list"},
	{"rewind", 0, IN, 0, OP_NONE, 0, f_rewind,
	 "rewind file(s)"},
	{"rm", 1, 1, 0, OP_NONE, 0, f_rm,
	 "remove a file"},
	{"root", 2, 3, 0, OP_NONE, 0, f_root,
	 "value a taken to the b'th root within accuracy c"},
	{"round", 1, 3, 0, OP_NONE, 0, f_round,
	 "round value a to b number of decimal places"},
	{"rsearch", 2, 4, 0, OP_NONE, 0, f_rsearch,
	 "reverse search matrix or list for value b\n\t\t    starting at index c"},
	{"runtime", 0, 0, 0, OP_NONE, f_runtime, 0,
	 "user mode cpu time in seconds"},
	{"scale", 2, 2, 0, OP_SCALE, 0, 0,
	 "scale value up or down by a power of two"},
	{"scan", 1, IN, FA, OP_NONE, 0, f_scan,
	 "scan standard input for assignment to one or more variables"},
	{"scanf", 2, IN, FA, OP_NONE, 0, f_scanf,
	 "formatted scan of standard input for assignment to variables"},
	{"search", 2, 4, 0, OP_NONE, 0, f_search,
	 "search matrix or list for value b starting\n\t\t    at index c"},
	{"sec", 1, 2, FE, OP_NONE, 0, f_sec,
	 "sec of a within accuracy b"},
	{"sech", 1, 2, FE, OP_NONE, 0, f_sech,
	 "hyperbolic secant of a within accuracy b"},
	{"seed", 0, 0, 0, OP_NONE, f_seed, 0,
	 "return a 64 bit seed for a psuedo-random generator"},
	{"segment", 2, 3, 0, OP_NONE, 0, f_segment,
	 "specified segment of specified list"},
	{"select", 2, 2, 0, OP_NONE, 0, f_select,
	 "form sublist of selected elements from list"},
	{"sgn", 1, 1, 0, OP_SGN, qsign, 0,
	 "sign of value (-1, 0, 1)"},
	{"sha", 0, IN, 0, OP_NONE, 0, f_sha,
	 "old Secure Hash Algorithm (SHS FIPS Pub 180)"},
	{"sha1", 0, IN, 0, OP_NONE, 0, f_sha1,
	 "Secure Hash Algorithm (SHS-1 FIPS Pub 180-1)"},
	{"sin", 1, 2, 0, OP_NONE, 0, f_sin,
	 "sine of value a within accuracy b"},
	{"sinh", 1, 2, FE, OP_NONE, 0, f_sinh,
	 "hyperbolic sine of a within accuracy b"},
	{"size", 1, 1, 0, OP_NONE, 0, f_size,
	 "total number of elements in value"},
	{"sizeof", 1, 1, 0, OP_NONE, 0, f_sizeof,
	 "number of bytes in memory storage for value"},
	{"sleep", 0, 1, 0, OP_NONE, 0, f_sleep,
	 "suspend operatioo for a seconds"},
	{"sort", 1, 1, 0, OP_NONE, 0, f_sort,
	 "sort a copy of a matrix or list"},
	{"sqrt", 1, 3, 0, OP_NONE, 0, f_sqrt,
	 "square root of value a within accuracy b"},
	{"srand", 0, 1, 0, OP_NONE, 0, f_srand,
	 "seed the rand() function"},
	{"srandom", 0, 4, 0, OP_NONE, 0, f_srandom,
	 "seed the random() function"},
	{"ssq", 1, IN, 0, OP_NONE, 0, f_ssq,
	 "sum of squares of values"},
	{"str", 1, 1, 0, OP_NONE, 0, f_str,
	 "simple value converted to string"},
	{"strcat", 1, IN, 0, OP_NONE, 0, f_strcat,
	 "concatenate strings together"},
	{"strcmp", 2, 2, 0, OP_NONE, 0, f_strcmp,
	 "compare two null-terminated strings"},
	{"strcpy", 2, 2, 0, OP_NONE, 0, f_strcpy,
	 "copy null-terminated string to string"},
	{"strerror", 0, 1, 0, OP_NONE, 0, f_strerror,
	 "string describing error type"},
	{"strlen", 1, 1, 0, OP_STRLEN, 0, 0,
	 "length of string"},
	{"strncmp", 3, 3, 0, OP_NONE, 0, f_strncmp,
	 "compare strings a, b to c characters"},
	{"strncpy", 3, 3, 0, OP_NONE, 0, f_strncpy,
	 "copy up to c characters from string to string"},
	{"strpos", 2, 2, 0, OP_NONE, 0, f_strpos,
	 "index of first occurrence of b in a"},
	{"strprintf", 1, IN, 0, OP_NONE, 0, f_strprintf,
	 "return formatted output as a string"},
	{"strscan", 2, IN, FA, OP_NONE, 0, f_strscan,
	 "scan a string for assignments to one or more variables"},
	{"strscanf", 2, IN, FA, OP_NONE, 0, f_strscanf,
	 "formatted scan of string for assignments to variables"},
	{"substr", 3, 3, 0, OP_NONE, 0, f_substr,
	 "substring of a from position b for c chars"},
	{"stacktrace", 0, 0, 0, OP_NONE, 0, f_stacktrace,
	 "string containing function call stack stack"},
	{"sum", 0, IN, 0, OP_NONE, 0, f_sum,
	 "sum of list or object sums and/or other terms"},
	{"swap", 2, 2, 0, OP_SWAP, 0, 0,
	 "swap values of variables a and b (can be dangerous)"},
	{"system", 1, 1, 0, OP_NONE, 0, f_system,
	 "call Unix command"},
	{"tail", 2, 2, 0, OP_NONE, 0, f_tail,
	 "retain list of specified number at tail of list"},
	{"tan", 1, 2, FE, OP_NONE, 0, f_tan,
	 "tangent of a within accuracy b"},
	{"tanh", 1, 2, FE, OP_NONE, 0, f_tanh,
	 "hyperbolic tangent of a within accuracy b"},
	{"test", 1, 1, 0, OP_TEST, 0, 0,
	 "test that value is nonzero"},
	{"time", 0, 0, 0, OP_NONE, f_time, 0,
	 "number of seconds since 00:00:00 1 Jan 1970 UTC"},
	{"trunc", 1, 2, 0, OP_NONE, f_trunc, 0,
	 "truncate a to b number of decimal places"},
	{"type", 1, 1, 0, OP_NONE, 0, f_type,
	 "string describing the type of a value"},
	{"ungetc", 2, 2, FO, OP_NONE, 0, f_ungetc,
	 "unget char read from file"},
	{"version", 0, 0, 0, OP_NONE, 0, f_version,
	 "calc version string"},
	{"xor", 1, IN, 0, OP_NONE, 0, f_xor,
	 "logical xor"},

	/* end of table */
	{NULL, 0, 0, 0, 0, 0, 0,
	 NULL}
};

#if defined(FUNCLIST)
/*ARGSUSED*/
int
main(int argc, char *argv[])
{
        CONST BuiltinFuncInfo *bp;       /* current function */

        printf("\nName\tArgs\tDescription\n\n");
        for (bp = builtins; bp->name; bp++) {
                printf("%-9s ", bp->name);
                if (bp->maxArgs == IN)
                        printf("%d+    ", bp->minArgs);
                else if (bp->minArgs == bp->maxArgs)
                        printf("%-6d", bp->minArgs);
                else
                        printf("%d-%-4d", bp->minArgs, bp->maxArgs);
                printf("%s\n", bp->desc);
        }
        printf("\n");
        return 0;       /* exit(0); */
}
#else /* FUNCLIST */
/*
 * Show the list of primitive built-in functions
 *
 * When FUNCLIST is defined, we are being compiled by rules from the help
 * sub-directory to form a program that will produce the main part of the
 * buiiltin help file.  These rules will convert the following function
 * name into main and remove the 'sed me out' line.
 *
 * See the builtin rule in the help/Makefile for details.
 */
void
ShowBuiltins(void)
{
	const BuiltinFuncInfo *	bp;	/* current function */

	printf("\nName\tArgs\tDescription\n\n");

	for (bp = builtins; bp->name; bp++)
	{
		if (bp->flags & FO)
			continue;

		printf("%-9s ", bp->name);

		if (bp->maxArgs == IN)
			printf("%d+    ", bp->minArgs);
		else if (bp->minArgs == bp->maxArgs)
			printf("%-6d", bp->minArgs);
		else
			printf("%d-%-4d", bp->minArgs, bp->maxArgs);

		printf("%s\n", bp->desc);
	}

	printf("\n");
}
#endif /* FUNCLIST */

#if !defined(FUNCLIST)


/*
 * Call a built-in function.
 * Arguments to the function are on the stack, but are not removed here.
 * Functions are either purely numeric, or else can take any value type.
 * This is slightly tricky in that our arguments can be addresses or
 * sub-addresses.
 *
 * given:
 *	index		index on where to scan in builtin table
 *	args		base of the arguments on the stack
 *	argCount	number of args
 */
VALUE
CallBuiltinFunction(long index, VALUE * args, int argCount)
{
	const BuiltinFuncInfo *	bp;	/* builtin function to be called */
	VALUE **	vpp;		/* pointer to current value address */
	int		i;		/* argument counter */
	int		subValueCount;	/* number of subvalues used */
	NUMBER *	numArgs[IN];	/* numeric arguments for function */
	VALUE *		valArgs[IN];	/* addresses of actual arguments */
	VALUE		result;		/* general result of function */

	/*
	 * Validate the built-in index.
	 */
	if ((unsigned long)index >=
	    (sizeof(builtins) / sizeof(builtins[0])) - 1)
	{
		math_error("Bad built-in function index");
		/*NOTREACHED*/
	}

	/*
	 * Get the pointer to the built-in information.
	 */
	bp = &builtins[index];

	/*
	 * Check the mininum and maximum argument counts.
	 */
	if (argCount < bp->minArgs)
	{
		math_error("Too few arguments for builtin function \"%s\"",
		    bp->name);
		/*NOTREACHED*/
	}

	if ((argCount > bp->maxArgs) || (argCount > IN))
	{
		math_error("Too many arguments for builtin function \"%s\"",
		    bp->name);
		/*NOTREACHED*/
	}

	/*
	 * Copy the argument value addresses into our temporary array.
	 * We need to do that since we are passed the raw addresses of
	 * variables, and we might need to derefence them to get their
	 * real values.  Only special functions preserve the addresses.
	 * Especially painful are sub-address values, in which case we
	 * have to extract their values and free them when we are done.
	 */
	vpp = valArgs;
	subValueCount = 0;

	for (i = argCount; i > 0; i--)
	{
		/*
		 * If this isn't an address value or if we want to pass
		 * address values onwards then store the real argument.
		 */
		if (!IS_ADDR_VALUE(args) || (bp->flags & FA))
			*vpp = args;
		else
		{
			/*
			 * We want to dereference addresses.
			 * Get the value from the full address or else
			 * from the sub-address.
			 */
			if (args->v_type == V_ADDR)
				*vpp = args->v_addr;
			else
			{
				/*
				 * Allocate room for the sub-value on
				 * the evaluation stack.
				 */
				*vpp = NewValue();

				/*
				 * Evaluate and save the subvalue on the stack.
				 */
				**vpp = LoadValueFromAddress(args);

				subValueCount++;
			}
		}

		args++;
		vpp++;
	}

	/*
	 * Handle general values if the function accepts them.
	 */
	if (bp->valueFunc)
	{
		vpp = valArgs;

		if ((bp->minArgs == 1) && (bp->maxArgs == 1))
			result = bp->valueFunc(vpp[0]);
		else if ((bp->minArgs == 2) && (bp->maxArgs == 2))
			result = bp->valueFunc(vpp[0], vpp[1]);
		else if ((bp->minArgs == 3) && (bp->maxArgs == 3))
			result = bp->valueFunc(vpp[0], vpp[1], vpp[2]);
		else
			result = bp->valueFunc(argCount, vpp, args);

		while (subValueCount-- > 0)
			PopValue();

		return result;
	}

	/*
	 * Function must be purely numeric, so handle that.
	 */
	vpp = valArgs;

	for (i = 0; i < argCount; i++)
	{
		if ((*vpp)->v_type != V_NUM)
		{
			math_error("Non-real argument for builtin function %s", bp->name);
			/*NOTREACHED*/
		}

		numArgs[i] = (*vpp)->v_num;
		vpp++;
	}

	result.v_type = V_NUM;

	if (!(bp->flags & FE) && (bp->minArgs != bp->maxArgs))
	{
		result.v_num = bp->numberFunc(argCount, numArgs);

		while (subValueCount-- > 0)
			PopValue();

		return result;
	}

	if ((bp->flags & FE) && (argCount < bp->maxArgs))
		numArgs[argCount++] = conf->epsilon;

	switch (argCount)
	{
		case 0:
			result.v_num = bp->numberFunc();
			break;

		case 1:
			result.v_num = bp->numberFunc(numArgs[0]);
			break;

		case 2:
			result.v_num = bp->numberFunc(numArgs[0], numArgs[1]);
			break;

		case 3:
			result.v_num = bp->numberFunc(numArgs[0], numArgs[1], numArgs[2]);
			break;

		default:
			math_error("Bad builtin function call");
			/*NOTREACHED*/
	}

	/*
	 * Remove any sub-values we used.
	 */
	while (subValueCount-- > 0)
		PopValue();

	/*
	 * Return the result.
	 */
	return result;
}


/*
 * Return the index of a built-in function given its name.
 * Returns minus one if the name is not known.
 */
int
GetBuiltinFunction(const char * name)
{
	const BuiltinFuncInfo *	bp;

	for (bp = builtins; bp->name; bp++)
	{
		if ((*name == *bp->name) && (strcmp(name, bp->name) == 0))
			return (bp - builtins);
	}

	return -1;
}


/*
 * Given the index of a built-in function, return its name.
 */
const char *
BuiltinName(long index)
{
	if ((unsigned long)index >=
	    (sizeof(builtins) / sizeof(builtins[0])) - 1)
	{
		return "";
	}

	return builtins[index].name;
}


/*
 * Given the index of a built-in function, and the number of arguments seen,
 * determine if the number of arguments are legal.  This routine is called
 * during parsing time.
 */
void
BuiltinCheck(long index, int count)
{
	const BuiltinFuncInfo *	bp;

	if ((unsigned long)index >=
	    (sizeof(builtins) / sizeof(builtins[0])) - 1)
	{
		math_error("Unknown built in index");
		/*NOTREACHED*/
	}

	bp = &builtins[index];

	if (count < bp->minArgs)
	{
		ScanError(T_NULL,
			"Too few arguments for builtin function \"%s\"",
			bp->name);
	}


	if (count > bp->maxArgs)
	{
		ScanError(T_NULL,
			"Too many arguments for builtin function \"%s\"",
			bp->name);
	}
}


/*
 * Return the opcode for a built-in function that can be used to avoid
 * the function call at all.  The special value of OP_NONE is returned
 * if there is no opcode for the built-in function.
 */
int
BuiltinOpcode(long index)
{
	if ((unsigned long)index >=
		(sizeof(builtins) / sizeof(builtins[0])) - 1)
	{
		return OP_NONE;
	}

	return builtins[index].opcode;
}

#endif /* FUNCLIST */
