/*
 * Edit a file using the user's prefered editor.
 * The files can either be permanent files or else temporary files
 * which are to be deleted when the calculator exits.
 * A list of all the edited files is kept to make repeated editing of
 * those files convenient.
 *
 * Copyright (C) 2009  David I. Bell
 *
 * Calc is open software; you can redistribute it and/or modify it under
 * the terms of the version 2.1 of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Calc is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU Lesser General
 * Public License for more details.
 *
 * A copy of version 2.1 of the GNU Lesser General Public License is
 * distributed with calc under the filename COPYING-LGPL.  You should have
 * received a copy with calc; if not, write to Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 */
#include "have_unistd.h"
#if defined(HAVE_UNISTD_H)
#include <unistd.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "calc.h"


/*
 * The temporary file name template which is appended to the temporary
 * directory path .  This name MUST end in six X's for the use of mkstemp.
 */
#define	TEMP_FILE_TEMPLATE	"/calc-XXXXXX"


/*
 * Structure used for holding files that have been edited.
 */
typedef	struct EditFile	EditFile;

struct	EditFile
{
	EditFile *	nextFile;		
	BOOL		isTemporary;
	long		id;
	char		fileName[1];
};


/*
 * Local data.
 */
static	long		nextFileId;
static	long		latestFileId = -1;
static	EditFile *	fileListHead = 0;
static	EditFile *	fileListTail = 0;
static	const char *	tempDir = "/tmp";


/*
 * Local routines.
 */
static	EditFile *	AllocEditFileEntry(const char *, BOOL);
static	EditFile *	FindEditFileEntryById(long);
static	EditFile *	FindEditFileEntryByName(const char *);
static	void		FreeEditFileEntry(EditFile *);


/*
 * Allocate a new temporary edit file entry and return its id.
 * Eventually the FreeEditFileEntry routine should be called to
 * free the entry and delete the temporary file.
 */
long
AllocTempEditFile(void)
{
	EditFile *	editFile;
	int		fd;
	mode_t		oldUmask;

	/*
	 * Find the temporary directory from the environment variables.
	 */
	tempDir = getenv("CALC_TMPDIR");

	if (tempDir == 0)
		tempDir = getenv("TMPDIR");

	if (tempDir == 0)
		tempDir = TMPDIR;

	/*
	 * Allocate an edit file entry in the temp directory.
	 */
	editFile = AllocEditFileEntry(tempDir, TRUE);

	/*
	 * Create a temporary file with restrictive permissions.
	 * Note: This will modify the template file name within the
	 * edit file entry to be the actual name of the file.
	 */
	oldUmask = umask(0177);

	fd = mkstemp(editFile->fileName);

	(void) umask(oldUmask);

	if (fd < 0)
		math_error("Failed to create temporary file in %s", tempDir);

	(void) close(fd);

	/*
	 * Return the edit file.
	 */
	latestFileId = editFile->id;

	return editFile->id;
}


/*
 * Allocate (if necessary) an edit file entry for the specified
 * file name and return its id.  If the file name is already known
 * then the previously allocated edit file entry is used.
 */
long
AllocPermEditFile(const char * fileName)
{
	EditFile *	editFile;

	editFile = FindEditFileEntryByName(fileName);

	if (editFile == 0)
		editFile = AllocEditFileEntry(fileName, FALSE);

	latestFileId = editFile->id;

	return editFile->id;
}


/*
 * Free the edit file entry having the specified id.
 * If the edit file is a temporary file, the file is deleted.
 * This does nothing if the id is not in use.
 */
void
FreeEditFile(long id)
{
	EditFile *	editFile;

	editFile = FindEditFileEntryById(id);

	if (editFile != 0)
		FreeEditFileEntry(editFile);
}


/*
 * Run the user's selected editor on the file specified by the
 * given edit file id.
 *
 * The editor specified by the CALC_EDITOR or EDITOR environment
 * variables is run on the file.
 *
 * After editing is completed, the length of the file is returned,
 * or -1 if the specified id is not in use or if the corresponding
 * file could not be edited due to some problem.
 */
long
EditFileById(long id)
{
	EditFile *	editFile;

	editFile = FindEditFileEntryById(id);

	if (editFile == 0)
		return -1;

	latestFileId = id;

	return EditFileByName(editFile->fileName);
}


/*
 * Run the user's selected editor on the specified file name.
 * The file is not limited to the set of file names contained in
 * the current edit file entries.
 *
 * The editor specified by the CALC_EDITOR or EDITOR environment
 * variables is run on the file.
 *
 * After editing is completed, the length of the file is returned,
 * or -1 if the file could not be edited due to some problem.
 */
long
EditFileByName(const char * fileName)
{
	const char *	editorName;
	char		command[256];
	struct stat	statBuf;

	/*
	 * Make sure there is a file name.
	 */
	if (fileName == 0)
		return -1;

	/*
	 * Get the editor program name to be run.
	 */
	editorName = getenv("CALC_EDITOR");

	if (editorName == 0)
		editorName = getenv("EDITOR");

	if (editorName == 0)
		editorName = EDITOR;

	/*
	 * TODO: Sanity check the editor name to prevent problems.
	 */

	/*
	 * Build the command line containing the editor and the file name.
	 */
	strcpy(command, editorName);
	strcat(command, " ");
	strcat(command, fileName);

	/*
	 * Run the command.
	 */
	if (system(command) < 0)
		return -1;

	/*
	 * Return the length of the file or -1 if the stat fails.
	 */
	if (stat(fileName, &statBuf) < 0)
		return -1;

	return statBuf.st_size;
}


/*
 * Return the name of the edit file that has the specified id.
 * The returned filename is only valid while the edit structure remains.
 * Returns NULL if the id is not in use.
 */
const char *
GetEditFileName(long id)
{
	EditFile *	editFile;

	editFile = FindEditFileEntryById(id);

	if (editFile == 0)
		return 0;

	return editFile->fileName;
}


/*
 * Return whether or not the edit file with the specified id is a
 * temporary file.  Returns FALSE if the id is not in use.
 */
BOOL
IsEditFileTemporary(long id)
{
	EditFile *	editFile;

	editFile = FindEditFileEntryById(id);

	if (editFile == 0)
		return 0;

	return editFile->isTemporary;
}


/*
 * Return the latest file id that was allocated.
 * If there was none or it had been invalidated, then -1 is returned.
 */
long
GetLatestEditFile(void)
{
	return latestFileId;
}


/*
 * Allocate a new edit file entry for the specified file name.
 * If the file is a temporary file name, then the specified file name
 * is assumed to be the temporary directory path, and the template
 * file name is appended to that path.
 */
static EditFile *
AllocEditFileEntry(const char * fileName, BOOL isTemporary)
{
	EditFile *	editFile;
	int		editSize;

	/*
	 * Calculate the size of the edit structure we need.
	 */
	editSize = sizeof(EditFile) + strlen(fileName);

	if (isTemporary)
		editSize += sizeof(TEMP_FILE_TEMPLATE);

	/*
	 * Allocate the edit file structure of the appropriate size.
	 */
	editFile = (EditFile *) malloc(editSize);

	if (editFile == 0)
		math_error("Cannot allocate EditFile");

	/*
	 * Initialize most of the fields.
	 */
	editFile->nextFile = 0;
	editFile->isTemporary = isTemporary;
	editFile->id = ++nextFileId;

	/*
	 * Copy in the file name.
	 */
	strcpy(editFile->fileName, fileName);

	if (isTemporary)
		strcat(editFile->fileName, TEMP_FILE_TEMPLATE);

	/*
	 * Link the structure into the list.
	 */
	if (fileListHead == 0)
		fileListHead = editFile;

	if (fileListTail != 0)
		fileListTail->nextFile = editFile;

	fileListTail = editFile;

	/*
	 * Return the structure.
	 */
	return editFile;
}


/*
 * Free the specified edit file entry.
 * This accepts a NULL pointer.
 */
static void
FreeEditFileEntry(EditFile * editFile)
{
	EditFile *	previousFile;
	BOOL		failed;

	/*
	 * Ignore null pointers.
	 */
	if (editFile == 0)
		return;

	/*
	 * If this was the latest file id then clear it.
	 */
	if (editFile->id == latestFileId)
		latestFileId = -1;

	/*
	 * Find the previous edit file entry to this one.
	 */
	previousFile = fileListHead;

	while ((previousFile != 0) && (previousFile->nextFile != editFile))
		previousFile = previousFile->nextFile;

	/*
	 * Unlink the entry from the list if it was found.
	 */
	if (previousFile != 0)
		previousFile->nextFile = editFile->nextFile;
	else if (editFile == fileListHead)
		fileListHead = editFile->nextFile;

	if (fileListTail == editFile)
		fileListTail = previousFile;

	/*
	 * If the file is a temporary file then try to delete it.
	 * Ignore any errors about the file not existing.
	 */
	failed = (editFile->isTemporary &&
		(unlink(editFile->fileName) < 0) && (errno != ENOENT));

	free(editFile);

	/*
	 * Complain if the file delete failed.
	 */
	if (failed)
		math_error("Could not delete temporary file");
}


/*
 * Find the edit file entry having the specified id.
 * Returns 0 if the id is not in use.
 */
static EditFile *
FindEditFileEntryById(long id)
{
	EditFile *	editFile;

	editFile = fileListHead;

	while ((editFile != 0) && (editFile->id != id))
		editFile = editFile->nextFile;

	return editFile;
}


/*
 * Find the edit file entry having the specified file name.
 * Returns 0 if none of the edit file entries have the specified name.
 */
static EditFile *
FindEditFileEntryByName(const char * fileName)
{
	EditFile *	editFile;

	editFile = fileListHead;

	while ((editFile != 0) && (strcmp(fileName, editFile->fileName) != 0))
		editFile = editFile->nextFile;

	return editFile;
}


/*
 * Delete all of the temporary files which we have created.
 * This is used when the calculator is about to exit.
 */
void
DeleteTempEditFiles(void)
{
	EditFile *	editFile;
	BOOL		retry;

	retry = TRUE;

	while (retry)
	{
		retry = FALSE;

		editFile = fileListHead;

		while ((editFile != 0) && !editFile->isTemporary)
			editFile = editFile->nextFile;

		if (editFile != 0)
		{
			FreeEditFileEntry(editFile);
			retry = TRUE;
		}
	}
}


/*
 * Show the list of files that have been edited.
 */
void
ShowEditFiles(void)
{
	EditFile *	editFile;
	char *		cp;
	char		flags[4];

	editFile = fileListHead;

	if (editFile == 0)
	{
		math_str("No files have been edited\n");

		return;
	}

	math_chr('\n');
	math_str(" Id    Flag  Filename\n");
	math_str(" ----  ----  --------\n");

	while (editFile != 0)
	{
		cp = flags;

		if (editFile->isTemporary)
			*cp++ = 'T';

		if (editFile->id == latestFileId)
			*cp++ = '*';

		*cp = '\0';

		math_fmt("% -5d  %-4s  ", editFile->id, flags);

		math_str(editFile->fileName);
		math_chr('\n');

		editFile = editFile->nextFile;
	}

	math_str("\n('T' - temporary, '*' - last edited)\n\n");
}
