/*
 * filefunc - file I/O value routines callable by users
 * This contains the functions and methods for files.
 * The low-level file I/O routines themselves are in file.c.
 *
 * Copyright (C) 1999  David I. Bell and Landon Curt Noll
 *
 * Primary author:  David I. Bell
 *
 * Calc is open software; you can redistribute it and/or modify it under
 * the terms of the version 2.1 of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Calc is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU Lesser General
 * Public License for more details.
 *
 * A copy of version 2.1 of the GNU Lesser General Public License is
 * distributed with calc under the filename COPYING-LGPL.  You should have
 * received a copy with calc; if not, write to Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * @(#) $Revision: 29.7 $
 * @(#) $Id: file.c,v 29.7 2001/04/10 22:06:46 chongo Exp $
 * @(#) $Source: /usr/local/src/cmd/calc/RCS/file.c,v $
 *
 * Under source code control:	1991/07/20 00:21:56
 * File existed as early as:	1991
 *
 * chongo <was here> /\oo/\	http://www.isthe.com/chongo/
 * Share and enjoy!  :-)	http://www.isthe.com/chongo/tech/comp/calc/
 */


#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <errno.h>
#include "have_unistd.h"
#if defined(HAVE_UNISTD_H)
# include <unistd.h>
#endif /* HAVE_UNISTD_H */
#include <ctype.h>
#include "calc.h"
#include "longbits.h"
#include "have_fpos.h"
#include "have_fpos_pos.h"
#include "fposval.h"
#include "file.h"
#include "calcerr.h"
#include "method.h"
#include "builtinfunc.h"

#if defined(_WIN32)
# include <io.h>
#endif


/*
 * Local routines.
 */
static int filescan(FILEID id, int count, VALUE ** vals);


/*
 * --------------------------------------------------------------
 *			FUNCTIONS FOR FILES
 * --------------------------------------------------------------
 */


/*
 * Determine access rights to a file.
 */
VALUE
f_access(int count, VALUE ** vals)
{
	NUMBER *q;
	int m;
	char *s, *fname;
	VALUE result;
	long i;

	/* initialize VALUE */
	result.v_type = V_NULL;
	result.v_subtype = V_NOSUBTYPE;

	errno = 0;
	if (vals[0]->v_type != V_STR)
		return error_value(E_ACCESS1);
	fname = vals[0]->v_str->s_str;
	m = 0;
	if (count == 2) {
		switch (vals[1]->v_type) {
			case V_NUM:
				q = vals[1]->v_num;
				if (qisfrac(q) || qisneg(q))
					return error_value(E_ACCESS2);
				m = (int)(q->num.v[0] & 7);
				break;
			case V_STR:
				s = vals[1]->v_str->s_str;
				i = (long)strlen(s);
				while (i-- > 0) {
					switch (*s++) {
					case 'r': m |= 4; break;
					case 'w': m |= 2; break;
					case 'x': m |= 1; break;
					default: return error_value(E_ACCESS2);
					}
				}
				break;
			case V_NULL:
				break;
			default:
				return error_value(E_ACCESS2);
		}
	}
	i = access(fname, m);
	if (i)
		return error_value(errno);
	return result;
}


/*
 * Remove one or more files.
 */
VALUE
f_rm(int count, VALUE ** vals)
{
	int	force;	/* TRUE -> -f was given as 1st arg */
	int	i;
	int	j;
	VALUE	result;

	/* initialize VALUE */
	result.v_subtype = V_NOSUBTYPE;

	/*
	 * firewall
	 */
	if (!allow_write)
		return error_value(E_WRPERM);

	/*
	 * check on each arg
	 */
	for (i=0; i < count; ++i) {
		if (vals[i]->v_type != V_STR)
			return error_value(E_RM1);
		if (vals[i]->v_str->s_str[0] == '\0')
			return error_value(E_RM1);
	}

	/*
	 * look for a leading -f option
	 */
	force = (strcmp(vals[0]->v_str->s_str, "-f") == 0);
	if (force) {
		--count;
		++vals;
	}

	/*
	 * remove file(s)
	 */
	for (i=0; i < count; ++i) {
		j = remove(vals[i]->v_str->s_str);
		if (!force && j < 0)
			return error_value(errno);
	}
	result.v_type = V_NULL;
	result.v_subtype = V_NOSUBTYPE;
	return result;
}


/*
 * Open a file.
 */
VALUE
f_fopen(VALUE * v1, VALUE * v2)
{
	FILEID		id;
	const char *	mode;
	VALUE		result;

	/* initialize VALUE */
	result.v_subtype = V_NOSUBTYPE;

	/* check for a valid mode [rwa][b+\0][b+\0] */
	if (v1->v_type != V_STR || v2->v_type != V_STR)
		return error_value(E_FOPEN1);
	mode = v2->v_str->s_str;
	if ((*mode != 'r') && (*mode != 'w') && (*mode != 'a'))
		return error_value(E_FOPEN2);
	if (mode[1] != '\0') {
		if (mode[1] != '+' && mode[1] != 'b')
			return error_value(E_FOPEN2);
		if (mode[2] != '\0') {
			if ((mode[2] != '+' && mode[2] != 'b') ||
			    mode[1] == mode[2])
				return error_value(E_FOPEN2);
			if (mode[3] != '\0')
				return error_value(E_FOPEN2);
		}
	}

	/* try to open */
	errno = 0;
	id = openid(v1->v_str->s_str, v2->v_str->s_str);
	if (id == FILEID_NONE)
		return error_value(errno);
	if (id < 0)
		return error_value(-id);
	result.v_type = V_FILE;
	result.v_file = id;
	return result;
}


/*
 * Re-open a file in another mode.
 */
VALUE
f_freopen(int count, VALUE ** vals)
{
	return m_FileReopen(0, vals[0], 0, vals + 1, count - 1);
}


/*
 * Close one or more files.
 * This is ugly since it can do multiple files or all files.
 */
VALUE
f_fclose(int count, VALUE ** vals)
{
	VALUE result;
	VALUE *vp;
	int n, i=0;

	/* initialize VALUE */
	result.v_subtype = V_NOSUBTYPE;

	errno = 0;
	if (count == 0) {
		i = closeall();
	} else {
		for (n = 0; n < count; n++) {
			vp = vals[n];
			if (vp->v_type != V_FILE)
				return error_value(E_FCLOSE1);
		}
		for (n = 0; n < count; n++) {
			vp = vals[n];
			i = closeid(vp->v_file);
			if (i < 0)
				return error_value(E_REWIND2);
		}
	}
	if (i < 0)
		return error_value(errno);
	result.v_type = V_NULL;
	return result;
}


/*
 * Return a reference to opened files.
 */
VALUE
f_files(int count, VALUE ** vals)
{
	VALUE	result;

	/* initialize VALUE */
	result.v_subtype = V_NOSUBTYPE;

	if (count == 0)
	{
		result.v_type = V_NUM;
		result.v_num = itoq((long) MAXFILES);

		return result;
	}

	if ((vals[0]->v_type != V_NUM) || qisfrac(vals[0]->v_num))
		return error_value(E_FILES);

	result.v_type = V_NULL;
	result.v_file = indexid(qtoi(vals[0]->v_num));

	if (result.v_file != FILEID_NONE)
		result.v_type = V_FILE;

	return result;
}


/*
 * Function to return whether a file is at end of file.
 */
VALUE
f_feof(VALUE * vp)
{
	return m_FileEof(0, vp, 0, 0, 0);
}


/*
 * Function to return whether a file has had an error.
 */
VALUE
f_ferror(VALUE * vp)
{
	return m_FileError(0, vp, 0, 0, 0);
}


/*
 * Function to flush a list of files.
 * This is ugly since it can do multiple files or all files.
 */
VALUE
f_fflush(int count, VALUE ** vals)
{
	VALUE result;
	int i, n;

	/* initialize VALUE */
	result.v_subtype = V_NOSUBTYPE;

	i = 0;
	errno = 0;
	if (count == 0) {
		i = flushall();
	} else {
		for (n = 0; n < count; n++) {
			if (vals[n]->v_type != V_FILE)
				return error_value(E_FFLUSH);
		}
		for (n = 0; n < count; n++) {
			i |= flushid(vals[n]->v_file);
		}
	}
	if (i == EOF)
		return error_value(errno);
	result.v_type = V_NULL;
	return result;
}


/*
 * Function to return file position of a file.
 */
VALUE
f_ftell(VALUE * vp)
{
	return m_FileTell(0, vp, 0, 0, 0);
}


VALUE
f_fseek(int count, VALUE **vals)
{
	return m_FileSeek(0, vals[0], 0, vals + 1, count - 1);
}


VALUE
f_rewind(int count, VALUE **vals)
{
	VALUE result;
	int n;

	/* initialize VALUE */
	result.v_subtype = V_NOSUBTYPE;

	if (count == 0) {
		rewindall();

	} else {
		for (n = 0; n < count; n++) {
			if (vals[n]->v_type != V_FILE)
				return error_value(E_REWIND1);
		}
		for (n = 0; n < count; n++) {
			if (rewindid(vals[n]->v_file) != 0) {
				return error_value(E_REWIND2);
			}
		}
	}
	result.v_type = V_NULL;
	return result;
}


VALUE
f_fsize(VALUE *vp)
{
	return m_FileSize(0, vp, 0, 0, 0);
}


VALUE
f_fprintf(int count, VALUE **vals)
{
	return m_FilePrintf(0, vals[0], 0, vals + 1, count - 1);
}


/*
 * Function to write a character to a file.
 */
VALUE
f_fputc(VALUE * v1, VALUE * v2)
{
	return m_FilePutc(0, v1, 0, &v2, 1);
}


/*
 * Function to write one or more strings to a file.
 */
VALUE
f_fputs(int count, VALUE ** vals)
{
	return m_FilePuts(0, vals[0], 0, vals + 1, count - 1);
}


VALUE
f_fputstr(int count, VALUE **vals)
{
	return m_FilePutstr(0, vals[0], 0, vals + 1, count - 1);
}


/*
 * Function to reread a character from a file.
 */
VALUE
f_ungetc(VALUE * v1, VALUE * v2)
{
	return m_FileUngetc(0, v1, 0, &v2, 1);
}


/*
 * Function to read a character from a file.
 */
VALUE
f_fgetc(VALUE * vp)
{
	return m_FileGetc(0, vp, 0, 0, 0);
}


/*
 * Function to read a line from a file.
 */
VALUE
f_fgetline(VALUE * vp)
{
	return m_FileGetline(0, vp, 0, 0, 0);
}


/*
 * Function to read a string from a file.
 */
VALUE
f_fgets(VALUE * vp)
{
	return m_FileGets(0, vp, 0, 0, 0);
}


/*
 * Function to read a string from a file.
 */
VALUE
f_fgetstr(VALUE * vp)
{
	return m_FileGetstr(0, vp, 0, 0, 0);
}


/*
 * Function to read a field from a file.
 */
VALUE
f_fgetfield(VALUE * vp)
{
	return m_FileGetfield(0, vp, 0, 0, 0);
}


VALUE
f_fscan(int count, VALUE **vals)
{
	VALUE *vp;
	VALUE result;
	int i;

	/* initialize VALUE */
	result.v_subtype = V_NOSUBTYPE;

	errno = 0;
	vp = *vals;
	if (vp->v_type == V_ADDR)
		vp = vp->v_addr;
	if (vp->v_type != V_FILE)
		return error_value(E_FSCAN1);

	i = filescan(vp->v_file, count - 1, vals + 1);

	if (i == EOF)
		return error_value(errno);
	if (i < 0)
		return error_value(E_FSCAN2);

	result.v_type = V_NUM;
	result.v_num = itoq((long) i);
	return result;
}


VALUE
f_scanf(int count, VALUE **vals)
{
	VALUE *vp;
	VALUE result;
	int i;

	/* initialize VALUE */
	result.v_subtype = V_NOSUBTYPE;

	vp = *vals;
	if (vp->v_type == V_ADDR)
		vp = vp->v_addr;
	if (vp->v_type != V_STR)
		return error_value(E_SCANF1);
	for (i = 1; i < count; i++) {
		if (vals[i]->v_type != V_ADDR)
			return error_value(E_SCANF2);
	}
	i = fscanfid(FILEID_STDIN, vp->v_str->s_str, count - 1, vals + 1);
	if (i < 0)
		return error_value(E_SCANF3);
	result.v_type = V_NUM;
	result.v_num = itoq((long) i);
	return result;
}


VALUE
f_fscanf(int count, VALUE **vals)
{
	VALUE *vp, *sp;
	VALUE result;
	int i;

	/* initialize VALUE */
	result.v_subtype = V_NOSUBTYPE;

	vp = *vals++;
	if (vp->v_type == V_ADDR)
		vp = vp->v_addr;
	if (vp->v_type != V_FILE)
		return error_value(E_FSCANF1);
	sp = *vals++;
	if (sp->v_type == V_ADDR)
		sp = sp->v_addr;
	if (sp->v_type != V_STR)
		return error_value(E_FSCANF2);
	for (i = 0; i < count - 2; i++) {
		if (vals[i]->v_type != V_ADDR)
			return error_value(E_FSCANF3);
	}
	i = fscanfid(vp->v_file, sp->v_str->s_str, count - 2, vals);
	if (i == EOF) {
		result.v_type = V_NULL;
		return result;
	}
	if (i < 0)
		return error_value(E_FSCANF4);
	result.v_type = V_NUM;
	result.v_num = itoq((long) i);
	return result;
}


/*
 * Utility routine for scanning a file.
 */
static int
filescan(FILEID id, int count, VALUE **vals)
{
	char	*str;
	int	i;
	int	n = 0;
	VALUE	val;
	VALUE	result;
	VALUE	*var;

	/* initialize VALUEs */
	val.v_type = V_STR;
	val.v_subtype = V_NOSUBTYPE;
	result.v_subtype = V_NOSUBTYPE;

	while (count-- > 0) {

		i = readid(id, 6, &str);

		if (i == EOF)
			break;
		if (i > 0)
			return EOF;
		n++;
		val.v_str = makenewstring(str);
		result = f_eval(&val);
		var = *vals++;
		if (var->v_type == V_ADDR) {
			var = var->v_addr;
			freevalue(var);
			*var = result;
		}
	}
	return n;
}


/*
 * --------------------------------------------------------------
 *			METHODS FOR FILES
 * --------------------------------------------------------------
 */
 
 
/*
 * Method to close a file.
 */
METHOD(m_FileClose)
{
	int	i;
	VALUE	result;

	if (value->v_type != V_FILE)
		return error_value(E_FEOF1);

	i = closeid(value->v_file);

	if (i < 0)
		return error_value(E_FEOF2);

	result.v_type = V_NUM;
	result.v_subtype = V_NOSUBTYPE;
	result.v_num = itoq((long) i);

	return result;
}


/*
 * Method to return whether a file is at end of file.
 */
METHOD(m_FileEof)
{
	int	i;
	VALUE	result;

	if (value->v_type != V_FILE)
		return error_value(E_FEOF1);

	i = eofid(value->v_file);

	if (i < 0)
		return error_value(E_FEOF2);

	result.v_type = V_NUM;
	result.v_subtype = V_NOSUBTYPE;
	result.v_num = itoq((long) i);

	return result;
}


/*
 * Method to return whether a file has had an error.
 */
METHOD(m_FileError)
{
	int	i;
	VALUE	result;

	if (value->v_type != V_FILE)
		return error_value(E_FERROR1);

	i = errorid(value->v_file);

	if (i < 0)
		return error_value(E_FERROR2);

	result.v_type = V_NUM;
	result.v_subtype = V_NOSUBTYPE;
	result.v_num = itoq((long) i);

	return result;
}


/*
 * Method to return the size of an opened file.
 */
METHOD(m_FileSize)
{
	int	i;
	ZVALUE	len;
	VALUE	result;

	if (value->v_type != V_FILE)
		return error_value(E_FSIZE1);

	i = getsize(value->v_file, &len);

	if (i == EOF)
		return error_value(errno);

	if (i)
		return error_value(E_FSIZE2);

	result.v_num = qalloc();
	result.v_num->num = len;
	result.v_type = V_NUM;
	result.v_subtype = V_NOSUBTYPE;

	return result;
}


/*
 * Method to flush buffered output to a file.
 */
METHOD(m_FileFlush)
{
	int	i;
	VALUE	result;

	if (value->v_type != V_FILE)
		return error_value(E_FERROR1);

	i = flushid(value->v_file);

	if (i < 0)
		return error_value(E_FERROR2);

	result.v_type = V_NUM;
	result.v_subtype = V_NOSUBTYPE;
	result.v_num = itoq((long) i);

	return result;
}


/*
 * Method to reread a character from a file.
 */
METHOD(m_FileUngetc)
{
	NUMBER *q;
	int	ch;
	int	i;
	VALUE	result;

	errno = 0;

	if (value->v_type != V_FILE)
		return error_value(E_UNGETC1);

	switch (args[0]->v_type)
	{
		case V_STR:
			ch = args[0]->v_str->s_str[0];
			break;

		case V_NUM:
			q = args[0]->v_num;

			if (!qisint(q))
				return error_value(E_UNGETC2);

			ch = qisneg(q) ? (int)(-q->num.v[0] & 0xff) :
					 (int)(q->num.v[0] & 0xff);

			break;

		default:
			return error_value(E_UNGETC2);
	}

	i = idungetc(value->v_file, ch);

	if (i == EOF)
		return error_value(errno);

	if (i == -2)
		return error_value(E_UNGETC3);

	result.v_type = V_NULL;
	result.v_subtype = V_NOSUBTYPE;

	return result;
}


/*
 * Method to get a character from a file.
 */
METHOD(m_FileGetc)
{
	int	ch;
	VALUE	result;

	if (value->v_type != V_FILE)
		return error_value(E_FGETC1);

	ch = getcharid(value->v_file);

	if (ch == -2)
		return error_value(E_FGETC2);

	result.v_type = V_NULL;
	result.v_subtype = V_NOSUBTYPE;

	if (ch != EOF)
	{
		result.v_type = V_STR;
		result.v_str = charstring(ch);
	}

	return result;
}


/*
 * Method to read a field from a file.
 */
METHOD(m_FileGetfield)
{
	char *	str;
	int	i;
	VALUE	result;

	if (value->v_type != V_FILE)
		return error_value(E_FGETFIELD1);

	i = readid(value->v_file, 14, &str);

	if (i > 0)
		return error_value(E_FGETFIELD2);

	result.v_type = V_NULL;
	result.v_subtype = V_NOSUBTYPE;

	if (i == 0)
	{
		result.v_type = V_STR;
		result.v_str = makestring(str);
	
	}

	return result;
}


/*
 * Method to read a line from a file.
 */
METHOD(m_FileGetline)
{
	char *	str;
	int	i;
	VALUE	result;

	if (value->v_type != V_FILE)
		return error_value(E_FGETLINE1);

	i = readid(value->v_file, 9, &str);

	if (i > 0)
		return error_value(E_FGETLINE2);

	result.v_type = V_NULL;
	result.v_subtype = V_NOSUBTYPE;

	if (i == 0)
	{
		result.v_type = V_STR;
		result.v_str = makestring(str);
	}

	return result;
}


METHOD(m_FileGets)
{
	char *	str;
	int	i;
	VALUE	result;

	if (value->v_type != V_FILE)
		return error_value(E_FGETS1);

	i = readid(value->v_file, 1, &str);

	if (i > 0)
		return error_value(E_FGETS2);

	result.v_type = V_NULL;
	result.v_subtype = V_NOSUBTYPE;

	if (i == 0)
	{
		result.v_type = V_STR;
		result.v_str = makestring(str);
	}

	return result;
}


METHOD(m_FileGetstr)
{
	char *	str;
	int	i;
	VALUE	result;

	if (value->v_type != V_FILE)
		return error_value(E_FGETSTR1);

	i = readid(value->v_file, 10, &str);

	if (i > 0)
		return error_value(E_FGETSTR2);

	result.v_type = V_NULL;
	result.v_subtype = V_NOSUBTYPE;

	if (i == 0)
	{
		result.v_type = V_STR;
		result.v_str = makestring(str);
	}

	return result;
}


/*
 * Method to print formatted output to a file.
 */
METHOD(m_FilePrintf)
{
	int	i;
	VALUE	result;

	if (value->v_type != V_FILE)
		return error_value(E_FPRINTF1);

	if (args[0]->v_type != V_STR)
		return error_value(E_FPRINTF2);

	i = idprintf(value->v_file, args[0]->v_str->s_str,
		argCount - 1, args + 1);

	if (i > 0)
		return error_value(E_FPRINTF3);

	result.v_type = V_NULL;
	result.v_subtype = V_NOSUBTYPE;

	return result;
}


/*
 * Method to write a character to a file.
 */
METHOD(m_FilePutc)
{
	NUMBER *q;
	int	ch;
	int	i;
	VALUE	result;

	if (value->v_type != V_FILE)
		return error_value(E_FPUTC1);

	switch (args[0]->v_type)
	{
		case V_STR:
			ch = args[0]->v_str->s_str[0];
			break;

		case V_NUM:
			q = args[0]->v_num;

			if (!qisint(q))
				return error_value(E_FPUTC2);

			ch = qisneg(q) ? (int)(-q->num.v[0] & 0xff) :
					 (int)(q->num.v[0] & 0xff);

			break;

		case V_NULL:
			ch = 0;
			break;

		default:
			return error_value(E_FPUTC2);
	}

	i = idfputc(value->v_file, ch);

	if (i > 0)
		return error_value(E_FPUTC3);

	result.v_type = V_NULL;
	result.v_subtype = V_NOSUBTYPE;

	return result;
}


/*
 * Method to write one or more strings (with newline) to a file.
 */
METHOD(m_FilePuts)
{
	int	index;
	int	err;
	VALUE	result;

	if (value->v_type != V_FILE)
		return error_value(E_FPUTS1);

	for (index = 0; index < argCount; index++)
	{
		if (args[index]->v_type != V_STR)
			return error_value(E_FPUTS2);
	}

	for (index = 0; index < argCount; index++)
	{
		err = idfputs(value->v_file, args[index]->v_str->s_str);

		if (err > 0)
			return error_value(E_FPUTS3);
	}

	result.v_type = V_NULL;
	result.v_subtype = V_NOSUBTYPE;

	return result;
}


/*
 * Method to write one or more strings to a file.
 */
METHOD(m_FilePutstr)
{
	int	i;
	int	err;
	VALUE	result;

	if (value->v_type != V_FILE)
		return error_value(E_FPUTSTR1);

	for (i = 0; i < argCount; i++)
	{
		if (args[i]->v_type != V_STR)
			return error_value(E_FPUTSTR2);
	}

	for (i = 0; i < argCount; i++)
	{
		err = idfputstr(value->v_file, args[i]->v_str->s_str);

		if (err > 0)
			return error_value(E_FPUTSTR3);
	}

	result.v_type = V_NULL;
	result.v_subtype = V_NOSUBTYPE;

	return result;
}


/*
 * Method to reopen a file.
 */
METHOD(m_FileReopen)
{
	FILEID		id;
	const char *	mode;
	VALUE		result;

	/*
	 * Check for a valid mode [rwa][b+\0][b+\0].
	 */
	if (value->v_type != V_FILE)
		return error_value(E_FREOPEN1);

	if (args[0]->v_type != V_STR)
		return error_value(E_FREOPEN2);

	mode = args[0]->v_str->s_str;

	if ((*mode != 'r') && (*mode != 'w') && (*mode != 'a'))
		return error_value(E_FREOPEN2);

	if (mode[1] != '\0')
	{
		if (mode[1] != '+' && mode[1] != 'b')
			return error_value(E_FREOPEN2);

		if (mode[2] != '\0')
		{
			if ((mode[2] != '+' && mode[2] != 'b') ||
			    mode[1] == mode[2])
			{
				return error_value(E_FOPEN2);
			}

			if (mode[3] != '\0')
				return error_value(E_FREOPEN2);
		}
	}

	/*
	 * Try to reopen.
	 */
	errno = 0;

	if (argCount == 2)
	{
		id = reopenid(value->v_file, mode, NULL);
	}
	else
	{
		if (args[1]->v_type != V_STR)
			return error_value(E_FREOPEN3);

		id = reopenid(value->v_file, mode,
			args[1]->v_str->s_str);
	}

	if (id == FILEID_NONE)
		return error_value(errno);

	result.v_type = V_NULL;
	result.v_subtype = V_NOSUBTYPE;

	return result;
}


/*
 * Method to scan input from a file.
 */
METHOD(m_FileScan)
{
	int	i;
	VALUE	result;

	errno = 0;

	if (value->v_type != V_FILE)
		return error_value(E_FSCAN1);

	i = filescan(value->v_file, argCount, args);

	if (i == EOF)
		return error_value(errno);

	if (i < 0)
		return error_value(E_FSCAN2);

	result.v_type = V_NUM;
	result.v_subtype = V_NOSUBTYPE;
	result.v_num = itoq((long) i);

	return result;
}


/*
 * Method to scan input from a file.
 */
METHOD(m_FileScanf)
{
	VALUE	result;

	result.v_type = V_NULL;
	result.v_subtype = V_NOSUBTYPE;

	return result;
}


/*
 * Method to seek within a file.
 */
METHOD(m_FileSeek)
{
	int	whence;
	int	i;
	VALUE	result;

	errno = 0;
	whence = 0;

	if (value->v_type != V_FILE)
		return error_value(E_FSEEK1);

	if (args[0]->v_type != V_NUM || qisfrac(args[0]->v_num))
		return error_value(E_FSEEK2);

	if (argCount > 1)
	{
		if (args[1]->v_type != V_NUM || qisfrac(args[1]->v_num) ||
			qisneg(args[1]->v_num))
		{
			return error_value(E_FSEEK2);
		}

		if (args[1]->v_num->num.len > 1)
			return error_value (E_FSEEK2);

		whence = (int)(unsigned int)(args[1]->v_num->num.v[0]);

		if (whence > 2)
			return error_value (E_FSEEK2);
	}

	i = fseekid(value->v_file, args[0]->v_num->num, whence);

	if (i == EOF)
		return error_value(errno);

	if (i < 0)
		return error_value(E_FSEEK3);

	result.v_type = V_NULL;
	result.v_subtype = V_NOSUBTYPE;

	return result;
}


/*
 * Method to return the file position of a file.
 */
METHOD(m_FileTell)
{
	int	i;
	ZVALUE	pos;
	VALUE	result;

	errno = 0;

	if (value->v_type != V_FILE)
		return error_value(E_FTELL1);

	i = ftellid(value->v_file, &pos);

	if (i < 0)
		return error_value(E_FTELL2);

	result.v_type = V_NUM;
	result.v_subtype = V_NOSUBTYPE;
	result.v_num = qalloc();
	result.v_num->num = pos;

	return result;
}
