/*
 * Copyright (C) 2003  David I. Bell
 *
 * Calc is open software; you can redistribute it and/or modify it under
 * the terms of the version 2.1 of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Calc is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU Lesser General
 * Public License for more details.
 *
 * A copy of version 2.1 of the GNU Lesser General Public License is
 * distributed with calc under the filename COPYING-LGPL.  You should have
 * received a copy with calc; if not, write to Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * @(#) $Revision: 29.7 $
 * @(#) $Id: value.h,v 29.7 2001/06/08 21:00:58 chongo Exp $
 * @(#) $Source: /usr/local/src/cmd/calc/RCS/value.h,v $
 *
 * Under source code control:	1993/07/30 19:42:47
 * File existed as early as:	1993
 *
 * Share and enjoy!  :-)	http://www.isthe.com/chongo/tech/comp/calc/
 *
 * Module to provide a value which manipulates a function or method.
 */

#if defined(CALC_SRC)	/* if we are building from the calc source tree */
# include "win32dll.h"
# include "config.h"
# include "calcerr.h"
# include "hash.h"
# include "string.h"
# include "value.h"
# include "builtinfunc.h"
# include "userfunc.h"
# include "method.h"
#else
# include <calc/win32dll.h>
# include <calc/config.h>
# include <calc/calcerr.h>
# include <calc/hash.h>
# include <calc/string.h>
# include <calc/value.h>
# include <calc/builtinfunc.h>
# include <calc/userfunc.h>
# include <calc/method.h>
#endif


/*
 * Local routines.
 */
static	UserFunc *	GetUserFunc(Func * func);


/*
 * Allocate a Func structure for the specified function or method name.
 * A method name is distinguished by the fact that the name contains a
 * period.  The period separates a type name from a method name.
 * Example:  "fred" is a function, but "string.fred" is a method.
 */
Func *
AllocFunc(const char * name)
{
	Func *		func;
	char *		type;
	const char *	cp;
	int		size;
	long		index;
	long		valueType;
	long		valueId;

	/*
	 * Calculate the size of the structure to hold the name
	 * and allocate it.
	 */
	size = sizeof(Func) + strlen(name);

	func = (Func *) malloc(size);

	if (func == 0)
		math_error("Cannot allocate Func structure");

	/*
	 * Initialize the fields.
	 */
	func->linkCount = 1;
	func->isMethod = FALSE;
	func->isBuiltin = FALSE;
	func->isUserDefined = FALSE;
	func->isValidType = FALSE;
	func->valueType = METHOD_WILD_ID;
	func->valueId = METHOD_WILD_ID;
	func->index = -1;
	func->type = 0;

	/*
	 * See if the name contains a period.
	 * If not, then this is a function name.
	 * If so, then this is a type name and a method name.
	 */
	cp = strchr(name, '.');

	if (cp == 0)
		strcpy(func->name, name);
	else
	{
		strcpy(func->name, cp + 1);
		type = func->name + strlen(func->name) + 1;
		memcpy(type, name, cp - name);
		type[cp - name] = '\0';
		func->type = type;
		func->isMethod = TRUE;
	}

	/*
	 * See if this is a built-in function name.
	 */
	if (!func->isMethod)
	{
		index = GetBuiltinFunction(name);

		if (index >= 0)
		{
			func->index = index;
			func->isBuiltin = TRUE;
		}
	}

	/*
	 * See if this is a valid method name and whether it is built-in.
	 */
	if (func->isMethod && ValueTypeAndId(func->type, &valueType, &valueId))
	{
		func->isValidType = TRUE;
		func->valueType = valueType;
		func->valueId = valueId;

		index = FindMethodName(func->name);

		if ((index >= 0) &&
			IsBuiltinMethod(func->name, valueType, valueId))
		{
			func->index = index;
			func->isBuiltin = TRUE;
		}
	}

	return func;
}


/*
 * Free a function value.
 * This can accept a null pointer.
 */
void
FreeFunc(Func * func)
{
	/*
	 * If the Func is NULL then do nothing.
	 */
	if (func == 0)
		return;

	/*
	 * Decrement the link count, and if it is nonzero then
	 * do nothing else.
	 */
	if (--(func->linkCount) != 0)
		return;

	/*
	 * This was the last reference to the Error so it needs freeing.
	 * All we have to free is the structure itself.
	 */
	free(func);
}


/*
 * Copy an Func value.
 * This accepts a null pointer.
 * This just increments the link count since a Func value can't change
 * in an externally visible way once they are allocated.
 */
Func *
CopyFunc(Func * func)
{
	/*
	 * Do nothing if this function is NULL.
	 */
	if (func == 0)
		return 0;

	/*
	 * Increment the link count and return the same function.
	 */
	func->linkCount++;

	return func;
}


/*
 * Link an Func value.
 * This acceps a null pointer.
 */
Func *
LinkFunc(Func * func)
{
	/*
	 * Do nothing if this function is NULL.
	 */
	if (func == 0)
		return 0;

	/*
	 * Increment the link count and return the same function.
	 */
	func->linkCount++;

	return func;
}


/*
 * Return whether two Func structures are the same.
 * This accepts NULL arguments.
 */
BOOL
IsSameFunc(const Func * func1, const Func * func2)
{
	/*
	 * If they are the same address then they are the same.
	 */
	if (func1 == func2)
		return TRUE;

	/*
	 * If only one of them is NULL then they are different.
	 */
	if ((func1 != 0) != (func2 != 0))
		return FALSE;

	/*
	 * They are the same if they have the same type and name.
	 * 
	 */
	if ((func1->type != 0) != (func2->type != 0))
		return FALSE;

	if ((func1->type != 0) && (strcmp(func1->type, func2->type) != 0))
		return FALSE;

	return (strcmp(func1->name, func2->name) == 0);
}


/*
 * Get or create a Func instance from the indicated VALUE.
 * The value can either already contain a Func value which is copied,
 * or else it can contain a string which names a function or method.
 * Other types of value will generate an error.  If the checkDefined
 * flag is TRUE and the specified function or method is not defined
 * then an error will be generated.  The returned Func value should be
 * freed when its use is completed.
 */
Func *
GetFuncFromValue(const VALUE * vp, BOOL checkDefined)
{
	Func *	func;

	/*
	 * Switch on the type of value to create the Func.
	 */
	switch (vp->v_type)
	{
		case V_STR:
			func = AllocFunc(vp->v_str->s_str);
			break;

		case V_FUNC:
			func = CopyFunc(vp->v_func);
			break;

		default:
			math_error("Function or string value required");
			/*NOTREACHED*/
			func = 0;
			break;
	}

	/*
	 * If requested make sure the function is defined.
	 */
	if (checkDefined && !FuncIsDefined(func))
	{
		if (!func->isMethod)
		{
			math_error("Function \"%s\" is not defined", func->name);
			/*NOTREACHED*/
		}

		math_error("Method \"%s.%s\" is not defined", func->type,
			func->name);
		/*NOTREACHED*/
	}

	/*
	 * Return the function.
	 */
	return func;
}


/*
 * Print a Func value using the specified flags.
 */
void
PrintFunc(const Func * func, int flags)
{
	math_str("Func \"");

	if (func->isMethod)
	{
		math_str(func->type);
		math_chr('.');
	}

	math_str(func->name);
	math_chr('\"');
}


/*
 * Return the number of bytes used by a Func value.
 * The wantOverhead flag is set to include overhead bytes.
 */
long
GetFuncSize(const Func * func, BOOL wantOverhead)
{
	if (func == 0)
		return 0;

	return sizeof(Func) + strlen(func->name);
}


/*
 * Utility routine to get the user function for the Func value.
 * An error is generated if the function is not defined or is built-in.
 */
static UserFunc *
GetUserFunc(Func * func)
{
	if (func->isBuiltin)
	{
		math_error("Func \"%s\" is built-in", func->name);
		/*NOTREACHED*/
	}

	if (FuncIsUserFunction(func))
		return FindUserFunction(func->index);

	if (func->isMethod)
	{
		math_error("Method \"%s.%s\" is not defined",
			func->type, func->name);
		/*NOTREACHED*/
	}

	math_error("Function \"%s\" is not defined", func->name);
	/*NOTREACHED*/

	return 0;
}


/*
 * Return whether the Func represents a defined function or method.
 * The function can be built-in or user-defined.
 */
BOOL
FuncIsDefined(Func * func)
{
	return (func->isBuiltin ||
		FuncIsUserFunction(func) || FuncIsUserMethod(func));
}


/*
 * Return whether the func represents a method name.
 * This doesn't indicate whether or not the method exists.
 */
BOOL
FuncIsMethod(const Func * func)
{
	return func->isMethod;
}


/*
 * Return whether the Func represents a function name.
 * This doesn't indicate whether or not the function exists.
 */
BOOL
FuncIsFunction(const Func * func)
{
	return !func->isMethod;
}


/*
 * Return whether the Func represents a built-in function.
 */
BOOL
FuncIsBuiltinFunction(const Func * func)
{
	return (func->isBuiltin && !func->isMethod);
}


/*
 * Return whether the Func represents a built-in method.
 */
BOOL
FuncIsBuiltinMethod(const Func * func)
{
	return (func->isBuiltin && func->isMethod);
}


/*
 * Return whether the Func represents a user-defined method.
 * This fact and the index of the user-defined method is saved when
 * it is discovered.
 */
BOOL
FuncIsUserMethod(Func * func)
{
	long	valueType;
	long	valueId;
	long	index;

	/*
	 * If this isn't a method name then it can't be a method.
	 */
	if (!func->isMethod)
		return FALSE;

	/*
	 * If we already thought that it is user-defined then
	 * return that fact.  We currently don't have a way to
	 * unregister methods.
	 */
	if (func->isUserDefined)
		return TRUE;

	/*
	 * If the type name wasn't valid then try to see if it is
	 * valid now.  (Maybe someone defined an object type after
	 * they created this function value.)
	 */
	if (!func->isValidType)
	{
		/*
		 * If the type is still bad then the user-defined
		 * method can't exist.
		 */
		if (!ValueTypeAndId(func->type, &valueType, &valueId))
			return FALSE;

		/*
		 * Update our information for later.
		 */
		func->isValidType = TRUE;
		func->valueType = valueType;
		func->valueId = valueId;
	}

	/*
	 * The type is correct.
	 * See if the method name is known for our value type.
	 * If so then remember that fact and return success.
	 */
	index = FindMethodName(func->name);

	if ((index >= 0) &&
		IsUserDefinedMethod(func->name, func->valueType, func->valueId))
	{
		func->index = index;
		func->isUserDefined = TRUE;

		return TRUE;
	}

	/*
	 * The user defined method does not exist.
	 */
	return FALSE;
}


/*
 * Return whether the Func represents a user-defined function.
 * This fact and the index of the user-defined function is saved when
 * it is discovered.
 */
BOOL
FuncIsUserFunction(Func * func)
{
	long	index;

	/*
	 * A built-in function or a method is never a user-defined
	 * function.
	 */
	if (func->isBuiltin || func->isMethod)
		return FALSE;

	/*
	 * If we already thought that it is user-defined then validate
	 * that fact since the user could have undefined it.
	 */
	if (func->isUserDefined)
	{
		/*
		 * If the function is still present then return TRUE.
		 */
		if (FindUserFunction(func->index) != 0)
			return TRUE;

		/*
		 * Clear the flag and return false since it has
		 * been undefined.
		 */
		func->isUserDefined = FALSE;

		return FALSE;
	}

	/*
	 * See if the function has been defined yet.
	 * If so then remember its index and return TRUE.
	 */
	index = GetUserFunction(func->name);

	if (index >= 0)
	{
		func->index = index;
		func->isUserDefined = TRUE;

		return TRUE;
	}

	/*
	 * The function is still not a user-defined function.
	 */
	return FALSE;
}


/*
 * Call the function or method referenced by a Func.
 * If this is a method then the first argument is the value to be
 * operated on.
 */
VALUE
FuncCall(Func * func, VALUE * args, int argCount)
{
	UserFunc *	userFunc;
	VALUE		dummyValue;

	/*
	 * If this is a built-in function then call it.
	 */
	if (!func->isMethod && func->isBuiltin)
		return CallBuiltinFunction(func->index, args, argCount);

	/*
	 * If this is a user-defined function then call it.
	 */
	if (!func->isMethod && FuncIsUserFunction(func))
	{
		userFunc = GetUserFunc(func);

		return Calculate(userFunc, 0, 0, args, argCount);
	}

	/*
	 * If this is a user defined or built-in method then call it.
	 */
	if (func->isMethod &&
		(func->isBuiltin || FuncIsUserMethod(func)))
	{
		/*
		 * Validate that the value matches the type that this
		 * Func value is defined to use.
		 */
		ValidateMethodValue(func->index, func->valueType,
			func->valueId, args);

		/*
		 * OK, make the call.
		 */
		return MethodCall(func->index, 0, args, args + 1, argCount - 1);
	}

	/*
	 * The function or method is not known.
	 */
	if (func->isMethod)
	{
		math_error("Method \"%s.%s\" is not defined",
			func->type, func->name);
		/*NOTREACHED*/
	}
	else
	{
		math_error("Function \"%s\" is not defined", func->name);
		/*NOTREACHED*/
	}

	/*
	 * To keep the compiler happy.
	 */
	dummyValue.v_type = V_NULL;

	return dummyValue;
}


/*
 * Method to get the name of the function.
 * For a method value, this only includes the part after the period.
 */
METHOD(m_FuncName)
{
	VALUE	result;

	if (value->v_type != V_FUNC)
		math_error("Non-Func argument");

	result.v_str = makenewstring(value->v_func->name);
	result.v_type = V_STR;
	result.v_subtype = V_NOSUBTYPE;

	return result;
}


/*
 * Method to get the type that the function acts on.
 * For a function, this returns null.
 */
METHOD(m_FuncType)
{
	VALUE	result;

	if (value->v_type != V_FUNC)
		math_error("Non-Func argument");

	result.v_type = V_NULL;
	result.v_subtype = V_NOSUBTYPE;

	if (value->v_func->type != 0)
	{
		result.v_str = makenewstring(value->v_func->type);
		result.v_type = V_STR;
	}

	return result;
}


/*
 * Method to determine whether the function or method is defined.
 */
METHOD(m_FuncIsDefined)
{
	int	r;
	VALUE	result;

	r = (value->v_func->isBuiltin || FuncIsUserFunction(value->v_func) ||
		FuncIsUserMethod(value->v_func));

	result.v_num = itoq(r);
	result.v_type = V_NUM;
	result.v_subtype = V_NOSUBTYPE;

	return result;
}


/*
 * Method to determine whether the function is built-in.
 */
METHOD(m_FuncIsBuiltin)
{
	VALUE	result;

	result.v_num = itoq(value->v_func->isBuiltin);
	result.v_type = V_NUM;
	result.v_subtype = V_NOSUBTYPE;

	return result;
}


/*
 * Method to determine whether the function is user-defined.
 */
METHOD(m_FuncIsUserDefined)
{
	VALUE	result;

	result.v_num = itoq(FuncIsUserFunction(value->v_func) ||
		FuncIsUserMethod(value->v_func));

	result.v_type = V_NUM;
	result.v_subtype = V_NOSUBTYPE;

	return result;
}


/*
 * Method to determine whether the function is a method.
 */
METHOD(m_FuncIsMethod)
{
	VALUE	result;

	result.v_num = itoq(value->v_func->isMethod);
	result.v_type = V_NUM;
	result.v_subtype = V_NOSUBTYPE;

	return result;
}


/*
 * Method to call a function or a method.
 * Our arguments are the arguments to the function or method.
 * For a method, the first argument is the value to operate on.
 */
METHOD(m_FuncCall)
{
	return FuncCall(value->v_func, rawArgs, argCount);
}


/*
 * Method to return the specified method's Func for this value.
 * This method works for any type of value.
 */
METHOD(m_WildFunc)
{
	const char *	type;
	const char *	name;
	VALUE		result;
	char		fullName[128];

	/*
	 * Make sure we have a string argument.
	 */
	if (argCount < 1)
	{
		math_error("Missing argument for \"func\" method");
		/*NOTREACHED*/
	}

	if (args[0]->v_type != V_STR)
	{
		math_error("Non-string argument for \"func\" method");
		/*NOTREACHED*/
	}

	/*
	 * Get the function name.
	 */
	name = args[0]->v_str->s_str;

	/*
	 * Get the type name of the this value.
	 */
	type = valuetype(value);

	/*
	 * Build a method name from these two.
	 */
	strcpy(fullName, type);
	strcat(fullName, ".");
	strcat(fullName, name);

	/*
	 * Allocate a function value for this name and return it.
	 */
	result.v_func = AllocFunc(fullName);
	result.v_type = V_FUNC;
	result.v_subtype = V_NOSUBTYPE;

	return result;
}


/*
 * Method to read or set the asserts checking flag for the function.
 * The old value of the assertion checking flag is returned.
 * If an argument is given then the flag is set to its value.
 */
METHOD(m_FuncAssertsEnabled)
{
	UserFunc *	userFunc;
	BOOL		checkFlag;
	VALUE		result;

	checkFlag = FALSE;

	/*
	 * If we have an argument then make sure it is an integer.
	 */
	if (argCount > 0)
	{
		if ((args[0]->v_type != V_NUM) || qisfrac(args[0]->v_num))
		{
			math_error("Non-integer argument for \"assertsEnabled\" method");
			/*NOTREACHED*/
		}

		checkFlag = !qiszero(args[0]->v_num);
	}

	/*
	 * Get the user function.
	 */
	userFunc = GetUserFunc(value->v_func);

	/*
	 * Build the return value.
	 */
	result.v_num = itoq(userFunc->assertsEnabled);
	result.v_type = V_NUM;
	result.v_subtype = V_NOSUBTYPE;

	/*
	 * If we were setting the flag then do so.
	 */
	if (argCount > 0)
		userFunc->assertsEnabled = checkFlag;

	/*
	 * Return the value.
	 */
	return result;
}
