/*
 * Copyright (C) 2009  David I. Bell
 *
 * Calc is open software; you can redistribute it and/or modify it under
 * the terms of the version 2.1 of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Calc is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU Lesser General
 * Public License for more details.
 *
 * A copy of version 2.1 of the GNU Lesser General Public License is
 * distributed with calc under the filename COPYING-LGPL.  You should have
 * received a copy with calc; if not, write to Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Module to allocate and free Node objects.
 * For convenience, this does not generate errors, but remembers them.
 * If an error is found while allocating a node, a special "bad" node
 * is returned.
 */

#include "have_unistd.h"
#if defined(HAVE_UNISTD_H)
#include <unistd.h>
#endif

#include <stdio.h>

#include "calc.h"
#include "token.h"
#include "symbol.h"
#include "string.h"

#include "node.h"
#include "nodeinfo.h"


#define	MAX_DUMP_DEPTH			100
#define	MAX_STATIC_LITERAL_NODES	8192


/*
 * Table of static literal nodes.
 * These are used for holding small integer values which can be shared.
 * These nodes are never modified once created and are never freed.
 */
static	Node *	StaticLiteralNodes[MAX_STATIC_LITERAL_NODES];


/*
 * A node which is returned on a node allocation error.
 */
static	Node	badNode;


/*
 * An error flag indicating that a node allocation error occurred.
 */
static	BOOL	hasNodeError;


/*
 * The next node to be returned by AllocNode if not 0.
 */
static	Node *	peekNode;


/*
 * Information used for dumping of nodes.
 */
static	FILE *	dumpFile;
static	int	dumpCurrentDepth;
static	int	dumpNodeCount;
static	int	dumpMaximumDepth;


/*
 * Local routines.
 */
static	void	DumpRecurse(const Node *);


/*
 * Reset the error flag for node allocation.
 */
void
ResetNodeError(void)
{
	hasNodeError = FALSE;
}


/*
 * Print an error message about a node error if it is not null,
 * set the node allocation error flag for later checking, and return
 * our special bad node.
 */
Node *
NodeError(const char * message)
{
	if (message != 0)
		ScanError(T_NULL, message);

	hasNodeError = TRUE;

	badNode.info = GetBadNodeInfo();
	badNode.kids = badNode.kidTable;

	return &badNode;
}


/*
 * Get whether or not there has been a node allocation error.
 */
BOOL
HasNodeError(void)
{
	return hasNodeError;
}


/*
 * Allocate a literal node which contains a specified integer value.
 * If the value is a small non-negative integer, this will return a
 * shared instance of a node from a static table.  Otherwise the node
 * will be newly allocated.  The FreeNode procedure can be called for
 * either type of result and will execute correctly.  Warning:  The
 * returned node cannot be modified since it might be shared.
 */
Node *
AllocLiteralNode(long value)
{
	Node * node;

	/*
	 * If the value is in range then use a static node
	 * otherwise allocate a new one.
	 */
	if ((value >= 0) && (value < MAX_STATIC_LITERAL_NODES))
	{
		/*
		 * Get the node from the static table.
		 */
		node = StaticLiteralNodes[value];

		/*
		 * If the node wasn't yet allocated, then initialize it.
		 */
		if (node == 0)
		{
			node = AllocNode(OP_INT_LITERAL);

			if (node == &badNode)
				return node;

			node->id = value;

			StaticLiteralNodes[value] = node;
		}

		/*
		 * Return the node.
		 */
		return node;
	}

	/*
	 * The value is out of our static range.
	 * Allocate a unique new node.
	 */
	node = AllocNode(OP_INT_LITERAL);

	/*
	 * Store the integer value into the id field.
	 */
	node->id = value;

	return node;
}


/*
 * Allocate an integer node for the specified integer value while
 * adding the number to the constant pool.
 */
Node *
AllocIntegerNode(long value)
{
	long		id;
	NUMBER *	q;

	/*
	 * Check for special cases.
	 */
	if (value == 0)
		return AllocNode(OP_ZERO);

	if (value == 1)
		return AllocNode(OP_ONE);

	if (value == -1)
		return AllocNode(OP_MINUS_ONE);

	/*
	 * Allocate a real number.
	 */
	q = itoq(value);

	/*
	 * Try to store the number in the constant table.
	 */
	id = addqconstant(q);

	if (id == 0)
	{
		math_error("Cannot allocate numeric constant");
		/*NOTREACHED*/
	}

	/*
	 * OK, allocate the numeric node.
	 */
	return AllocNodeWithId(OP_NUMBER, id);
}


/*
 * Routine to allocate a numeric opcode for the specified number,
 * while adding the number to the constant pool.  The number is
 * freed if it is not actually used.  Returns 0 if the node could
 * not be allocated.
 */
Node *
AllocNumberNode(NUMBER * q)
{
	long	id;

	/*
	 * Check for zero.
	 */
	if (qiszero(q))
	{
		qfree(q);

		return AllocNode(OP_ZERO);
	}

	/*
	 * Check for one.
	 */
	if (qisone(q))
	{
		qfree(q);

		return AllocNode(OP_ONE);
	}

	/*
	 * Check for minus one.
	 */
	if (qisnegone(q))
	{
		qfree(q);

		return AllocNode(OP_MINUS_ONE);
	}

	/*
	 * Try to store the number in the constant table.
	 */
	id = addqconstant(q);

	if (id == 0)
	{
		qfree(q);

		return 0;
	}

	/*
	 * OK, allocate the numeric node.
	 */
	return AllocNodeWithId(OP_NUMBER, id);
}


/*
 * Allocate a node for a unary operation opcode with one kid.
 * More kids can be added later if required by the opcode.
 */
Node *
AllocUnaryNode(int opcode, Node * kid)
{
	Node *	node;

	if (kid == 0)
		return NodeError("Null kid of unary node");

	node = AllocNode(opcode);

	if (node == &badNode)
		return node;

	node->count = 1;
	node->kids[0] = kid;

	return node;
}


/*
 * Allocate a node for a binary operation opcode with two kids.
 * More kids can be added later if required by the opcode.
 */
Node *
AllocBinaryNode(int opcode, Node * left, Node * right)
{
	Node *	node;

	if ((left == 0) || (right == 0))
		return NodeError("Null kid of binary node");

	node = AllocNode(opcode);

	if (node == &badNode)
		return node;

	node->count = 2;
	node->kids[0] = left;
	node->kids[1] = right;

	return node;
}


/*
 * Append another kid to the end of the kids table in the specified node.
 */
void
AppendKidToNode(Node * node, Node * kid)
{
	Node **	newkids;

	if ((node == 0) || (kid == 0))
	{
		ScanError(T_NULL, "Bad add kid to node");

		return;
	}

	if (node == &badNode)
		return;

	/*
	 * Reallocate the kid array if it is full.
	 */
	if ((node->count > 0) && ((node->count % NORMAL_KIDS) == 0))
	{
		newkids = (Node **)
			malloc((node->count + NORMAL_KIDS) * sizeof(Node *));

		if (newkids == 0)
		{
			ScanError(T_NULL, "Cannot grow node table");

			return;
		}

		memcpy((char *) newkids, (char *) node->kids,
			node->count * sizeof(Node *));

		if (node->kids != node->kidTable)
			free(node->kids);

		node->kids = newkids;
	}

	/*
	 * Add the new kid to the node.
	 */
	node->kids[node->count++] = kid;
}


/*
 * Allocate a node of the specified type with no kinds and with a
 * specified id value.
 */
Node *
AllocNodeWithId(int opcode, long id)
{
	Node *	node;

	node = AllocNode(opcode);

	if (node != &badNode)
		node->id = id;

	return node;
}


/*
 * Allocate a node of the specified type with no kids and with the
 * specified pointer data value.  The data should not point to data
 * which requires freeing unless FreeNode is updated to know about it.
 */
Node *
AllocNodeWithData(int opcode, const void * data)
{
	Node *	node;

	node = AllocNode(opcode);

	if (node == &badNode)
		return node;

	node->data = data;

	return node;
}


/*
 * Allocate a boolean node.
 * This is just a convenience function which returns a node
 * containing the OP_ZERO or OP_ONE opcodes.
 */
Node *
AllocBooleanNode(BOOL flag)
{
	return AllocNode(flag ? OP_ONE : OP_ZERO);
}


/*
 * Allocate a node of the specified type with no kids.
 */
Node *
AllocNode(int opcode)
{
	Node *			node;
	const NodeInfo *	nodeInfo;

	/*
	 * Find the node information for this opcode.
	 */
	nodeInfo = FindNodeInfo(opcode);

	/*
	 * If the opcode was not known then complain.
	 */
	if (nodeInfo == 0)
		return NodeError("Node info not found for opcode");

	/*
	 * Allocate the node.
	 * But first try to return the peek Node if it is set.
	 */
	node = peekNode;
	peekNode = 0;

	if (node == 0)
	{
		node = (Node *) malloc(sizeof(Node));

		if (node == 0)
			return NodeError("Cannot allocate Node");
	}

	/*
	 * Initialise the node.
	 */
	memset((char *) node, 0, sizeof(Node));

	node->info = nodeInfo;
	node->kids = node->kidTable;

	/*
	 * Return the node.
	 */
	return node;
}


/*
 * Peek at the next node structure that will be allocated.
 * This node will be the next one which is returned by AllocNode.
 * The returned node CANNOT be used for any purpose except for
 * remembering its address since nothing is filled out.
 * Returns 0 if the peek node cannot be allocated.
 */
Node *
PeekAllocNode(void)
{
	Node * node;

	/*
	 * If the peek node is already allocated then return it.
	 */
	if (peekNode)
		return peekNode;

	/*
	 * Allocate a new node.
	 */
	node = (Node *) malloc(sizeof(Node));

	if (node == 0)
		return 0;

	/*
	 * Initialize the node a bit just in case it is looked at.
	 */
	memset((char *) node, 0, sizeof(Node));

	node->info = GetBadNodeInfo();
	node->kids = node->kidTable;

	/*
	 * Save the peek node and return it.
	 */
	peekNode = node;

	return node;
}


/*
 * Free a node and all of its kids.
 * This is careful to not wrongly free a static literal node or
 * the special bad or nop nodes.
 */
void
FreeNode(Node * node)
{
	int	index;

	/*
	 * If the node is NULL or is the bad node then do nothing,
	 */
	if ((node == 0) || (node == &badNode))
		return;

	/*
	 * Free a possible NUMBER constant that the node used.
	 */
	if (((node->info->flags & NODE_FLAG_NUMBER_DATA) != 0) &&
		(node->id >= 0))
	{
		freeconstant(node->id);
	}

	/*
	 * Free a possible STRING constant that the node used.
	 */
	if (((node->info->flags & NODE_FLAG_STRING_DATA) != 0) &&
		(node->id >= 0))
	{
		freestringconstant(node->id);
	}

	/*
	 * Free all of the kids.
	 */
	for (index = 0; index < node->count; index++)
	{
		FreeNode(node->kids[index]);
		node->kids[index] = 0;
	}

	node->count = 0;

	/*
	 * If the kids table was allocated then free it.
	 */
	if (node->kids != node->kidTable)
	{
		free(node->kids);

		node->kids = node->kidTable;
	}

	/*
	 * If the peek node is this one then clear it.
	 */
	if (peekNode == node)
		peekNode = 0;

	/*
	 * If this is not a literal integer node or its value is
	 * not in range of the static table then free the node itself.
	 */
	if ((node->info->opcode != OP_INT_LITERAL) ||
		((node->id < 0) || (node->id > MAX_STATIC_LITERAL_NODES)))
	{
		free(node);
	}
}


/*
 * Return the integer value storaged in a special integer literal node.
 * An error is generated if the node is not valid.
 */
int
GetNodeLiteralInteger(const Node * node)
{
	if (node->info->opcode != OP_INT_LITERAL)
		math_error("Wrong node for literal integer value");

        return (int) node->id;          /* TODO: Check for overflow */ 
}


/*
 * Find the kid number of the specified kid node within another node.
 * Returns -1 if the kid node was not found.
 */
int
FindKidNodeIndex(const Node * node, const Node * kidNode)
{
	int	index;

	/*
	 * Look through the block's kids for a match.
	 */
	for (index = 0; index < node->count; index++)
	{
		if (node->kids[index] == kidNode)
			return index;
	}

	/*
	 * The target node was not found.
	 */
	return -1;
}


/*
 * Return whether or not the node is an assignment node.
 */
BOOL
NodeIsAssignment(const Node * node)
{
	return ((node->info->flags & NODE_FLAG_ASSIGNMENT) != 0);
}


/*
 * Return whether a node is considered a constant.
 */
BOOL
NodeIsConst(const Node * node)
{
	return node->info->isRepeatable(node);
}


/*       
 * Return whether a node is considered an rvalue.
 * An rvalue is one which cannot be assigned to.
 */
BOOL
NodeIsRValue(const Node * node)
{
	return !node->info->isLValue(node);
}


/*
 * Recursive routine to copy a node tree.
 */
Node *
CopyNodeTree(const Node * oldNode)
{
	Node *	node;
	Node *	kidNode;
	int	index;

	/*
	 * If the old node is null or the bad node then return that.
	 */
	if (oldNode == 0)
		return 0;

	if (oldNode == &badNode)
		return &badNode;

	/*
	 * If the node is a small literal integer node then
	 * allocate a new one specially.
	 */
	if (oldNode->info->opcode == OP_INT_LITERAL)
		return AllocLiteralNode(oldNode->id);

	/*
	 * Allocate a new copy of the node.
	 */
	node = (Node *) malloc(sizeof(Node));

	if (node == 0)
		math_error("No memory for node copy");

	/*
	 * Copy information from the old node.
	 */
	node->info = oldNode->info;
	node->id = oldNode->id;
	node->data = oldNode->data;
	node->lineNumber = oldNode->lineNumber;

	/*
	 * If the node contains a NUMBER data value with a number index
	 * defined then increment the link count for the number.
	 */
	if (((node->info->flags & NODE_FLAG_NUMBER_DATA) != 0) &&
		(oldNode->id >= 0))
	{
		LinkConstValue(oldNode->id);
	}

	/*
	 * If the node contains a STRING data value with a string
	 * index defined then increment the link count for the string.
	 */
	if (((node->info->flags & NODE_FLAG_STRING_DATA) != 0) &&
		(oldNode->id >= 0))
	{
		LinkStringConstant(oldNode->id);
	}

	/*
	 * Initialize the kid table, and the make copies of the kids.
	 */
	node->kids = node->kidTable;
	node->count = 0;

	for (index = 0; index < oldNode->count; index++)
	{
		kidNode = CopyNodeTree(oldNode->kids[index]);

		AppendKidToNode(node, kidNode);
	}

	/*
	 * Return the copied node.
	 */
	return node;
}


/*
 * Dump the node and its kids starting from the specified node to a
 * previously maximum depth (in case of node loops).  Each level of
 * nodes is indented by an extra couple of spaces.  The dumped
 * information includes the node type, the number of kids, and possibly
 * other node-specific information.  The information is written to the
 * supplied file handle.
 */
void
DumpNodeTree(FILE * file, const Node * node, const char * funcName,
	const char * message)
{
	dumpFile = file;
	dumpNodeCount = 0;
	dumpCurrentDepth = 0;
	dumpMaximumDepth = MAX_DUMP_DEPTH;

	if ((message == 0) || (*message == '\0'))
		fprintf(file, "------- Node Dump of \"%s\"\n", funcName);
	else
	{
		fprintf(file, "------- Node Dump of \"%s\" (%s)\n",
			funcName, message);
	}

	DumpRecurse(node);

	fprintf(file, "------- End Node Dump (%d node%s)\n",
		dumpNodeCount, ((dumpNodeCount == 1) ? "" : "s"));
}


/*
 * Dump just the specified node to the indicated file.
 * No extraneous output is produced.
 */
void
DumpNode(FILE * file, const Node * node)
{
	dumpFile = file;
	dumpNodeCount = 0;
	dumpCurrentDepth = 0;
	dumpMaximumDepth = 0;

	DumpRecurse(node);
}


/*
 * Recursive routine to dump and node and its kids.
 */
static void
DumpRecurse(const Node * node)
{
	const NodeInfo *	info;
	const char *		nodeText;
	int			depth;
	int			index;

	/*
	 * Increment the number of nodes dumped.
	 */
	dumpNodeCount++;

	/*
	 * Indent the description according to the current depth.
	 */
	for (depth = 0; depth < dumpCurrentDepth; depth++)
		fprintf(dumpFile, ".");

	/*
	 * If there is no node then complain.
	 */
	if (node == 0)
	{
		fprintf(dumpFile, "Node is NULL\n");

		return;
	}

	/*
	 * If there is no node info then complain.
	 */
	info = node->info;

	if (info == 0)
	{
		fprintf(dumpFile, "Node has NULL info\n");

		return;
	}

	/*
	 * Dump specific information describing the node.
	 */
	nodeText = info->dump(node);

	/*
	 * Dump out the node information.
	 */
	fprintf(dumpFile, "%s%s%s", info->name,
		((nodeText == 0) ? "" : ": "),
		((nodeText == 0) ? "" : nodeText));

	if (node->lineNumber > 0)
		fprintf(dumpFile, "  [line %d]", node->lineNumber);

	fputc('\n', dumpFile);

	/*
	 * If we haven't reached the maximum depth yet then
	 * call ourself for the kids of this node.
	 */
	if (dumpCurrentDepth < dumpMaximumDepth)
	{
		dumpCurrentDepth++;

		for (index = 0; index < node->count; index++)
			DumpRecurse(node->kids[index]);

		dumpCurrentDepth--;
	}
}


/*
 * Count the number of nodes in the specific node tree.
 * This calls ourself recursively to find the answer.
 */
long
CountNodeTree(const Node * node)
{
	int	index;
	long	count;

	if (node == 0)
		return 0;

	count = 1;

	for (index = 0; index < node->count; index++)
		count += CountNodeTree(node->kids[index]);

	return count;
}

/* END CODE */
