/*
 * obj - object handling primitives
 *
 * Copyright (C) 1999  David I. Bell
 *
 * Calc is open software; you can redistribute it and/or modify it under
 * the terms of the version 2.1 of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Calc is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU Lesser General
 * Public License for more details.
 *
 * A copy of version 2.1 of the GNU Lesser General Public License is
 * distributed with calc under the filename COPYING-LGPL.  You should have
 * received a copy with calc; if not, write to Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * @(#) $Revision: 29.2 $
 * @(#) $Id: obj.c,v 29.2 2000/06/07 14:02:13 chongo Exp $
 * @(#) $Source: /usr/local/src/cmd/calc/RCS/obj.c,v $
 *
 * Under source code control:	1990/02/15 01:48:19
 * File existed as early as:	before 1990
 *
 * Share and enjoy!  :-)	http://www.isthe.com/chongo/tech/comp/calc/
 */
/*
 * "Object" handling primitives.
 * This simply means that user-specified routines are called to perform
 * the indicated operations.  These can be operated on either by
 * procedural-style function or by object-oriented-style methods.
 * Some built-in methods also exist for some default handling.
 */

#include <stdio.h>
#include "calc.h"
#include "userfunc.h"
#include "symbol.h"
#include "string.h"
#include "method.h"


/*
 * Allocation chunk size for objects actions.
 */
#define OBJALLOC 16


/*
 * Types of values returned by calling object routines.
 */
#define A_VALUE 0	/* returns arbitrary value */
#define A_INT	1	/* returns integer value */
#define A_UNDEF 2	/* returns no value */
#define	A_THIS	3	/* returns THIS value */


/*
 * Default built-in method actions for objects.
 */
#define D_ERROR		0	/* generate method not found error */
#define D_PRINT		1	/* print element */
#define D_CMP		2	/* compare two values */
#define D_TEST		3	/* test value for nonzero */
#define D_POW		4	/* call generic power routine */
#define D_ONE		5	/* return number 1 */
#define D_INC		6	/* increment by one */
#define D_DEC		7	/* decrement by one */
#define D_SQUARE	8	/* square value */
#define D_VALUE		9	/* return value */
#define D_ASSIGN	10	/* assign value */
#define	D_NOP		11	/* do nothing */
#define	D_FALSE		12	/* return false value */


/*
 * Information about built-in object methods and functions.
 */
typedef	struct
{
	short		argCount;	/* number of arguments */
	BOOL		isBinary;	/* this is a binary operator */
	short		retVal;		/* type of return value */
	short		defaultAction;	/* special action on errors */
	const char *	name;		/* name of function to call */
	const char *	comment;	/* useful comment if any */
	long		methodId;	/* method id */
} ObjectInfo;


/*
 * Table of built-in methods and functions.
 * NOTE: The order of the elements in this table must match the OBJ_XXX
 * definitions in value.h.  The table must also end with an entry
 * containing NULL values.
 */
static	ObjectInfo	objectInfo[OBJ_MAXFUNC + 2] =
{
	{1, FALSE, A_UNDEF, D_PRINT,	"print","print value, default prints elements"},
	{1, FALSE, A_VALUE, D_ONE,	"one",	"multiplicative identity, default is 1"},
	{1, FALSE, A_INT,   D_TEST,	"test", "logical test (false,true => 0,1), default tests elements"},
	{2, TRUE,  A_VALUE, D_ERROR,	"add",	"sum"},
	{2, TRUE,  A_VALUE, D_ERROR,	"sub",	"difference"},
	{1, FALSE, A_VALUE, D_ERROR,	"neg",	"negative"},
	{2, TRUE,  A_VALUE, D_ERROR,	"mul",	"product"},
	{2, TRUE,  A_VALUE, D_ERROR,	"div",	"non-integral division"},
	{1, FALSE, A_VALUE, D_ERROR,	"inv",	"multiplicative inverse"},
	{2, FALSE, A_VALUE, D_ERROR,	"abs",	"absolute value within given error"},
	{1, FALSE, A_VALUE, D_ERROR,	"norm", "square of absolute value"},
	{1, FALSE, A_VALUE, D_ERROR,	"conj", "conjugate"},
	{2, FALSE, A_VALUE, D_POW,	"pow",	"integer power, default does multiply, square, inverse"},
	{1, FALSE, A_VALUE, D_ERROR,	"sgn",	"sign of value (-1, 0, 1)"},
	{2, FALSE, A_INT,   D_CMP,	"cmp",	"equality (equal,nonequal => 0,1), default tests elements"},
	{2, FALSE, A_VALUE, D_ERROR,	"rel",	"relative order, positive for >, etc."},
	{3, TRUE,  A_VALUE, D_ERROR,	"quo",	"integer quotient"},
	{3, TRUE,  A_VALUE, D_ERROR,	"mod",	"remainder of division"},
	{1, FALSE, A_VALUE, D_ERROR,	"int",	"integer part"},
	{1, FALSE, A_VALUE, D_ERROR,	"frac", "fractional part"},
	{1, FALSE, A_VALUE, D_INC,	"inc",	"increment, default adds 1"},
	{1, FALSE, A_VALUE, D_DEC,	"dec",	"decrement, default subtracts 1"},
	{1, FALSE, A_VALUE, D_SQUARE,	"square","default multiplies by itself"},
	{2, FALSE, A_VALUE, D_ERROR,	"scale", "multiply by power of 2"},
	{2, FALSE, A_VALUE, D_ERROR,	"shift", "shift left by n bits (right if negative)"},
	{3, FALSE, A_VALUE, D_ERROR,	"round", "round to given number of decimal places"},
	{3, FALSE, A_VALUE, D_ERROR,	"bround","round to given number of binary places"},
	{3, FALSE, A_VALUE, D_ERROR,	"root", "root of value within given error"},
	{3, FALSE, A_VALUE, D_ERROR,	"sqrt", "square root within given error"},
	{2, TRUE,  A_VALUE, D_ERROR,	"or",	"bitwise or"},
	{2, TRUE,  A_VALUE, D_ERROR,	"and",	"bitwise and"},
	{1, FALSE, A_VALUE, D_ERROR,	"not",	"logical not"},
	{1, FALSE, A_VALUE, D_ERROR,	"fact", "factorial or postfix !"},
	{1, FALSE, A_VALUE, D_VALUE,	"min",	"value for min(...)"},
	{1, FALSE, A_VALUE, D_VALUE,	"max",	"value for max(...)"},
	{1, FALSE, A_VALUE, D_VALUE,	"sum",	"value for sum(...)"},
	{2, FALSE, A_UNDEF, D_ASSIGN,	"assign", "assign, defaults to a = b"},
	{2, TRUE,  A_VALUE, D_ERROR,	"xor",	"value for binary ~"},
	{1, FALSE, A_VALUE, D_ERROR,	"comp", "value for unary ~"},
	{1, FALSE, A_VALUE, D_ERROR,	"content", "unary hash op"},
	{2, TRUE,  A_VALUE, D_ERROR,	"hashop",  "binary hash op"},
	{1, FALSE, A_VALUE, D_ERROR,	"backslash", "unary backslash op"},
	{2, TRUE,  A_VALUE, D_ERROR,	"setminus", "binary backslash op"},
	{1, FALSE, A_VALUE, D_ERROR,	"plus", "unary + op"},
	{1, FALSE, A_UNDEF, D_NOP,	"init",	"initialize value"},
	{2, FALSE, A_THIS,  D_FALSE,	"thisAdd",	"add value"},
	{2, FALSE, A_THIS,  D_FALSE,	"thisSubtract",	"subtract value"},
	{2, FALSE, A_THIS,  D_FALSE,	"thisMultiply",	"multiply by value"},
	{2, FALSE, A_THIS,  D_FALSE,	"thisDivide",	"divide by value"},
	{2, FALSE, A_THIS,  D_FALSE,	"thisQuotient",	"quotient by value"},
	{2, FALSE, A_THIS,  D_FALSE,	"thisModulo",	"modulo by value"},
	{2, FALSE, A_THIS,  D_FALSE,	"thisPower",	"power by value"},
	{2, FALSE, A_THIS,  D_FALSE,	"thisLeftShift","shift left by value"},
	{2, FALSE, A_THIS,  D_FALSE,    "thisRightShift", "shift right by value"},
	{2, FALSE, A_THIS,  D_FALSE,	"thisOr",	"bitwise or by value"},
	{2, FALSE, A_THIS,  D_FALSE,	"thisAnd",	"bitwise and by value"},
	{2, FALSE, A_THIS,  D_FALSE,	"thisXor",	"bitwise xor by value"},
	{2, FALSE, A_THIS,  D_FALSE,	"thisScale",	"scale value"},
	{0, FALSE, 0, 0, NULL}
};


/*
 * Other local data.
 */
static	STRINGHEAD		objectNames;	/* names of objects */
static	STRINGHEAD		elements;	/* element names for parts of objects */
static	OBJECTACTIONS **	objects; 	/* table of actions for objects */
static	long			maxObjectCount = 0;


/*
 * Local routines.
 */
static	UserFunc *	FindObjectFunction(const VALUE *, int);
static	const char *	ObjectMethodName(const char *);
static	VALUE		ObjectPower(VALUE * vp, NUMBER * q);
static	BOOL		ObjectTest(OBJECT * op);
static	BOOL		ObjectCompare(OBJECT * op1, OBJECT * op2);
static	void		ObjectPrint(const OBJECT * op);

static	METHOD(BuiltinObjectMethod);


/*
 * Register all of our method names to correspond with the old-style
 * procedural names.  This creates methods having names of the forms:
 *	opName
 * where name is the action name given in the objectInfo table.
 */
void
RegisterObjectMethods(void)
{
	ObjectInfo *	oip;
	const char *	methodName;
	long		action;
	int		argCount;

	/*
	 * Loop over all of the objectInfo table.
	 */
	for (oip = objectInfo; oip->name; oip++)
	{
		/*
		 * The method data to be passed back on a method call
		 * is the index into the table of the action which is
		 * the action code.
		 */
		action = (oip - objectInfo);

		/*
		 * Generate the name of the method.
		 */
		methodName = ObjectMethodName(oip->name);

		/*
		 * Calculate the required argument count for the method.
		 * This is usually one less than the number used by the
		 * corresponding procedural function since the first
		 * argument is taken for the value to be operated on.
		 * However, for a binary operation another argument is
		 * added back in to indicate the rightHandSide flag.
		 */
		argCount = oip->argCount;

		if (!oip->isBinary)
			argCount--;

		/*
		 * Register the method and remember its id in the
		 * objectInfo table for later use.
		 */
		oip->methodId = RegisterMethod(methodName, V_OBJ,
			METHOD_WILD_ID, argCount, argCount,
			BuiltinObjectMethod, action, 0);
	}
}


/*
 * Method to execute one of our built-in methods for objects.
 * These are only used if there is neither a user-defined method for
 * the action nor a user-defined procedural function for the action.
 * Therefore, this provides the default handling for all object types.
 * Note: Most of the default handling actions generate an error.
 * Since the built-in methods are only called from our own code we
 * don't have to check the arguments very much.
 */
static
METHOD(BuiltinObjectMethod)
{
	const ObjectInfo *	oip;
	long			action;
	BOOL			isRHS;
	VALUE *			RHSValue;
	VALUE *			v1;
	VALUE *			v2;
	VALUE			result;
	VALUE			tempValue;

	result.v_type = V_NULL;
	result.v_subtype = V_NOSUBTYPE;

	v1 = value;
	v2 = rawArgs;

	/*
	 * Get the built-in action value from the method info and
	 * range check it.
	 */
	action = methodInfo->methodData;

	if (((unsigned) action) > OBJ_MAXFUNC)
	{
		math_error("Illegal action %d for object call", action);
		/*NOTREACHED*/
	}

	/*
	 * Get the information about the action.
	 */
	oip = &objectInfo[action];

	/*
	 * If this is a binary operand then set the isRHS flag
	 * according to whether our object was used on the right hand
	 * or left hand side of the binary operand.  This may affect
	 * some of our handling.
	 */
	isRHS = FALSE;

	if (oip->isBinary)
	{
		RHSValue = &rawArgs[argCount - 1];

		if (RHSValue->v_type != V_NUM)
		{
			math_error("Bad RHS value for built-in method call");
			/*NOTREACHED*/
		}

		isRHS = !qiszero(RHSValue->v_num);
	}

	/*
	 * Switch on the default action.
	 */
	switch (oip->defaultAction)
	{
		case D_PRINT:
			ObjectPrint(v1->v_obj);
			break;

		case D_CMP:
			result.v_type = V_INT;

			if (v1->v_type != v2->v_type)
				result.v_int = 1;
			else
			{
				result.v_int = ObjectCompare(v1->v_obj,
					v2->v_obj);
			}

			break;

		case D_TEST:
			result.v_type = V_INT;
			result.v_int = ObjectTest(v1->v_obj);
			break;

		case D_POW:
			if (v2->v_type != V_NUM)
			{
				math_error("Non-real power");
				/*NOTREACHED*/
			}

			result = ObjectPower(v1, v2->v_num);
			break;

		case D_ONE:
			result.v_type = V_NUM;
			result.v_num = qlink(&_qone_);
			break;

		case D_INC:
			tempValue.v_type = V_NUM;
			tempValue.v_subtype = V_NOSUBTYPE;
			tempValue.v_num = &_qone_;
			result = ObjectCall(OBJ_ADD, v1, &tempValue, NULL_VALUE);
			break;

		case D_DEC:
			tempValue.v_type = V_NUM;
			tempValue.v_subtype = V_NOSUBTYPE;
			tempValue.v_num = &_qone_;
			result = ObjectCall(OBJ_SUB, v1, &tempValue, NULL_VALUE);
			break;

		case D_SQUARE:
			result = ObjectCall(OBJ_MUL, v1, v1, NULL_VALUE);
			break;

		case D_VALUE:
			copyvalue(v1, &result);
			break;

		case D_ASSIGN:
			if (isRHS)
			{
				v1 = rawArgs;
				v2 = value;
			}
			copyvalue(v2, &tempValue);
			tempValue.v_subtype = v1->v_subtype;
			freevalue(v1);
			*v1 = tempValue;
			result.v_type = V_NULL;
			break;

		case D_ERROR:
			math_error("No \"%s\" method for object type \"%s\"",
				ObjectMethodName(oip->name),
				namestr(&objectNames,
					value->v_obj->o_actions->oa_index));

			/*NOTREACHED*/

		case D_NOP:
			break;

		case D_FALSE:
			/*
			 * Return a false value.
			 */
			result.v_type = V_INT;
			result.v_subtype = V_NOSUBTYPE;
			result.v_int = FALSE;

			break;

		default:
			math_error("Unknown default action");
			/*NOTREACHED*/
	}

	return result;
}


/*
 * Construct a string for the indicated base action name.
 * The string is in static memory and so is only valid until the next call.
 * The name is of the forms:
 *	opName
 */
static const char *
ObjectMethodName(const char * actionName)
{
	char *		cp;
	static	char	name[32];

	cp = name;
	strcpy(cp, "op");
	cp += strlen(cp);
	*cp++ = *actionName++ - 'a' + 'A';
	strcpy(cp, actionName);

	return name;
}


/*
 * Show all the routine names available for objects.
 */
void
showobjfuncs(void)
{
	const ObjectInfo *	oip;

	printf("\nThe following object routines are definable.\n");
	printf("Note: xx represents the actual object type name.\n\n");
	printf("Name	Args	Comments\n");

	for (oip = objectInfo; oip->name; oip++)
	{
		printf("xx_%-8s %d	%s\n", oip->name, oip->argCount,
			oip->comment ? oip->comment : "");
	}

	printf("\n");
}


/*
 * Show all the method names available for objects.
 */
void
ShowObjectMethods(void)
{
	const ObjectInfo *	oip;
	int			argCount;

	printf("\nThe following implicit object methods are definable.\n");
	printf("A * means the method is a binary operator where the last\n");
	printf("argument is 1 if the object is the second operand.\n\n");

	printf("Name       Args   Comments\n");

	for (oip = objectInfo; oip->name; oip++)
	{
		argCount = oip->argCount - 1;

		if (oip->isBinary)
			argCount++;

		printf("%-11s %d %c   %s\n", ObjectMethodName(oip->name),
			argCount, (oip->isBinary ? '*' : ' '),
			(oip->comment ? oip->comment : ""));
	}

	printf("\n");
}


/*
 * Call the appropriate user-defined routine to handle an object action.
 * This can be an old-style user-defined function, a new-style user-defined
 * method call, or a new-style built-in method call.  These are searched
 * for in the following order:
 *	User-defined method call
 *	User-defined procedural call
 *	Built-in method call
 * If none of these are found then an error is generated.
 * The object to act on can be either the first or the second value.
 * This fact can be passed to the calling method as its last argument.
 * Returns the value that the routine returned.
 */
VALUE
ObjectCall(int action, VALUE * v1, VALUE * v2, VALUE * v3)
{
	const OBJECTACTIONS *	oap;	/* object to call for */
	const UserFunc *	func;	/* function to call */
	const ObjectInfo *	oip;		/* information about action */
	const MethodInfo *	methodInfo;	/* method info */
	int		argCount;	/* number of expected arguments */
	long		methodId;	/* method Id to call */
	long		valueId;	/* value Id for method */
	long		index;		/* index of function (negative if undefined) */
	BOOL		isBinary;	/* action is a binary operator */
	BOOL		rightHandSide;	/* object is on right hand side */
	VALUE *		args;		/* the arguments for the call */
	VALUE *		originalV1;	/* original v1 argument */
	VALUE *		objectValue;	/* value being operated on */
	VALUE		value;		/* return value */
	VALUE		nullValue;	/* null value */

	/*
	 * Get the information for this action.
	 */
	if (((unsigned) action) > OBJ_MAXFUNC)
	{
		math_error("Illegal action for object call");
		/*NOTREACHED*/
	}

	oip = &objectInfo[action];

	methodId = oip->methodId;
	argCount = oip->argCount;
	isBinary = oip->isBinary;

	/*
	 * Default the argument values if they were null.
	 */
	nullValue.v_type = V_NULL;
	nullValue.v_subtype = V_NOSUBTYPE;

	originalV1 = v1;

	if (v1 == 0)
		v1 = &nullValue;

	if (v2 == 0)
		v2 = &nullValue;

	if (v3 == 0)
		v3 = &nullValue;

	/*
	 * Determine which value is the object being operated on.
	 * This sets whether the object is on the left hand side or
	 * the right hand side of a binary operator.  For a unary
	 * operator, it's always on the left hand side.
	 */
	rightHandSide = FALSE;
	objectValue = 0;

	if ((v1->v_type == V_OBJ) || !isBinary)
	{
		objectValue = v1;
	}
	else if ((argCount > 1) && (v2->v_type == V_OBJ))
	{
		objectValue = v2;
		rightHandSide = TRUE;
	}

	/*
	 * Make sure we got an object argument.
	 */
	if ((objectValue == 0) || (objectValue->v_type != V_OBJ))
	{
		math_error("Object call called with non-object");
		/*NOTREACHED*/
	}

	/*
	 * Get the object actions for the object being operated on
	 * and the value id for the object.
	 */
	oap = objectValue->v_obj->o_actions;
	valueId = oap->oa_index;

	/*
	 * Find the method information for this type of object.
	 * This should always exist since we registered the wildcard
	 * object having this method Id.  (The user may have defined
	 * his own method for the object too in which case this will
	 * return information about that.)
	 */
	methodInfo = FindMethodInfo(methodId, 0, V_OBJ, valueId);

	if (methodInfo == 0)
	{
		math_error("Missing method info for object");
		/*NOTREACHED*/
	}

	/*
	 * If there is no user-defined method for the action then try
	 * to find an old-style user-defined function for this object.
	 * We don't try to find one if the user also defined a method.
	 */
	func = 0;

	if (methodInfo->func == 0)
		func = FindObjectFunction(objectValue, action);

	/*
	 * Call the user-defined function if it was found or
	 * else call the user-defined or built-in method.
	 */
	if (func != 0)
	{
		/*
		 * We have a user-defined function to call.
		 * Push linked arguments onto the evaluation stack in
		 * preparation for the call.  In this way, the call is
		 * still efficient like an address pointer would be, but
		 * the parameter can be assigned to without affecting
		 * the actual value.
		 */
		PushLinkedValue(v1);
		args = TopValue();

		if (argCount > 1)
			PushLinkedValue(v2);

		if (argCount > 2)
			PushLinkedValue(v3);

		/*
		 * Call the user-defined function.
		 */
		value = Calculate(func, 0, 0, args, argCount);

		/*
		 * Free the pushed arguments.
		 */
		PopValue();

		if (argCount > 1)
			PopValue();

		if (argCount > 2)
			PopValue();
	}
	else
	{
		/*
		 * We are calling a user-defined or built-in method.
		 * There is one less argument for the method call since
		 * one of the arguments was the object value which we
		 * pass in explicitly.
		 */
		argCount--;

		/*
		 * Copy the arguments into the array.  Note that the
		 * first argument to the method varies depending on
		 * whether the object was the first or second operand.
		 */
		if (rightHandSide)
			PushLinkedValue(v1);
		else
			PushLinkedValue(v2);

		args = TopValue();

		if (argCount > 1)
			PushLinkedValue(v3);

		/*
		 * If this is a binary operator then append an argument to
		 * the end which indicates whether or not the object was
		 * the right hand side operand.  The method can then check
		 * this value to handle that case.
		 */
		if (isBinary)
		{
			PushBoolean(rightHandSide);
			argCount++;
		}

		/*
		 * Call the method.
		 */
		value = MethodCallUsingMethodInfo(methodInfo, 0, objectValue,
			args, argCount);

		/*
		 * Free the arguments.
		 */
		PopValue();

		if (argCount > 1)
			PopValue();

		if (argCount > 2)
			PopValue();

		/*
		 * If this was a built-in method call then the return
		 * value type is already known to be correct.
		 */
		if (methodInfo->func == 0)
			return value;
	}

	/*
	 * The result was calculated by a user-defined function or method.
	 * Handle the expected return type.
	 */
	switch (oip->retVal)
	{
		case A_VALUE:
			/*
			 * The calculated value is what we want.
			 */
			break;

		case A_UNDEF:
			/*
			 * We always want to return the undefined value.
			 * Free any returned value.
			 */
			freevalue(&value);

			value.v_type = V_NULL;
			value.v_subtype = V_NOSUBTYPE;

			break;

		case A_INT:
			/*
			 * A small integer value is expected.
			 */
			if ((value.v_type != V_NUM) || qisfrac(value.v_num))
			{
				math_error("Integer return value required");
				/*NOTREACHED*/
			}

			index = qtoi(value.v_num);
			qfree(value.v_num);

			value.v_type = V_INT;
			value.v_subtype = V_NOSUBTYPE;
			value.v_int = index;

			break;

		case A_THIS:
			/*
			 * The THIS object value was expected.
			 */
			if (value.v_type == V_ADDR)
				value = value.v_addr[0];

			if ((v1 == 0) || (value.v_type != v1->v_type) ||
				(value.v_type != V_OBJ) ||
				(value.v_obj != v1->v_obj))
			{
				math_error("The \"this\" value was not returned");
				/*NOTREACHED*/
			}

			/*
			 * Return success value.
			 */
			value.v_type = V_INT;
			value.v_subtype = V_NOSUBTYPE;
			value.v_int = TRUE;

			break;

		default:
			/*
			 * There is an unknown return type expected.
			 */
			math_error("Bad object return action %d", oip->retVal);
			/*NOTREACHED*/
	}

	/*
	 * Return the calculated value.
	 */
	return value;
}


/*
 * Perform a object call on an object which operates on its first value.
 * Returns TRUE depending on whether or not the operation was successful.
 * This depends on whether the appropriate method or function was defined.
 */
BOOL
ThisObjectCall(int action, VALUE * v1, VALUE * v2, VALUE * v3)
{
	VALUE	result;

	result = ObjectCall(action, v1, v2, v3);

	if (result.v_type != V_INT)
	{
		math_error("Bad result value \"%s\" from ThisObjectCall",
			valuetype(&result));
		/*NOTREACHED*/
	}

	return (result.v_int != 0);
}


/*
 * Find the user-defined function (if any) for the specified action
 * on this type of object.  Returns 0 if there is no such function.
 */
static UserFunc *
FindObjectFunction(const VALUE * objectValue, int action)
{
	OBJECTACTIONS *		oap;
	const ObjectInfo *	oip;
	long			index;
	char			name[128];

	/*
	 * Validate the arguments.
	 */
	if ((objectValue == 0) || (objectValue->v_type != V_OBJ))
		math_error("Non-object for FindObjectFunction");

	if (((unsigned) action) > OBJ_MAXFUNC)
	{
		math_error("Illegal action for FindObjectFunction");
		/*NOTREACHED*/
	}

	/*
	 * Get pointers to the required data structures.
	 */
	oip = &objectInfo[action];
	oap = objectValue->v_obj->o_actions;

	/*
	 * If we don't know the index for the function name yet then
	 * build the name and add it into the function table, and
	 * save the index for the next call.
	 */
	index = oap->oa_indices[action];

	if (index < 0)
	{
		strcpy(name, namestr(&objectNames, oap->oa_index));
		strcat(name, "_");
		strcat(name, oip->name);
		index = AddUserFunction(name);
		oap->oa_indices[action] = index;

		if (index < 0)
			return 0;
	}

	/*
	 * Return the user-defined function if it exists, or 0.
	 */
	return FindUserFunction(index);
}


/*
 * Print the elements of an object in short and unambiguous format.
 * This is the default routine if the user's is not defined.
 *
 * given:
 *	op		object being printed
 */
static void
ObjectPrint(const OBJECT * op)
{
	int count;		/* number of elements */
	int i;			/* index */

	count = op->o_actions->oa_count;

	math_fmt("obj %s {", namestr(&objectNames, op->o_actions->oa_index));

	for (i = 0; i < count; i++)
	{
		if (i)
			math_str(", ");

		printvalue(&op->o_table[i], PRINT_SHORT | PRINT_UNAMBIG);
	}

	math_chr('}');
}


/*
 * Test an object for being "nonzero".
 * This is the default routine if the user's is not defined.
 * Returns TRUE if any of the elements are "nonzero".
 */
static BOOL
ObjectTest(OBJECT * op)
{
	int	i;

	i = op->o_actions->oa_count;

	while (--i >= 0)
	{
		if (testvalue(&op->o_table[i]))
			return TRUE;
	}

	return FALSE;
}


/*
 * Compare two objects for equality, returning TRUE if they differ.
 * This is the default routine if the user's is not defined.
 * For equality, all elements must be equal.
 */
static BOOL
ObjectCompare(OBJECT * op1, OBJECT * op2)
{
	int	i;

	if (op1->o_actions != op2->o_actions)
		return TRUE;

	i = op1->o_actions->oa_count;

	while (--i >= 0)
	{
		if (comparevalue(&op1->o_table[i], &op2->o_table[i]))
			return TRUE;
	}

	return FALSE;
}


/*
 * Raise an object to an integral power.
 * This is the default routine if the user's is not defined.
 * Negative powers mean the positive power of the inverse.
 * Zero means the multiplicative identity.
 *
 * given:
 *	vp		value to be powered
 *	q		power to raise number to
 */
static VALUE
ObjectPower(VALUE * vp, NUMBER * q)
{
	long	power;		/* power to raise to */
	FULL	bit;		/* current bit value */
	VALUE	res;
	VALUE	tmp;

	if (qisfrac(q))
	{
		math_error("Raising object to non-integral power");
		/*NOTREACHED*/
	}

	if (zge31b(q->num))
	{
		math_error("Raising object to very large power");
		/*NOTREACHED*/
	}

	power = ztolong(q->num);

	if (qisneg(q))
		power = -power;

	/*
	 * Handle some low powers specially
	 */
	if ((power <= 2) && (power >= -2))
	{
		switch ((int) power)
		{
			case 0:
				return ObjectCall(OBJ_ONE, vp, NULL_VALUE,
					NULL_VALUE);
	
			case 1:
				res.v_obj = objcopy(vp->v_obj);
				res.v_type = V_OBJ;
				res.v_subtype = V_NOSUBTYPE;
				return res;

			case -1:
				return ObjectCall(OBJ_INV, vp, NULL_VALUE,
					NULL_VALUE);
	
			case 2:
				return ObjectCall(OBJ_SQUARE, vp, NULL_VALUE,
					NULL_VALUE);
		}
	}

	if (power < 0)
		power = -power;

	/*
	 * Compute the power by squaring and multiplying.
	 * This uses the left to right method of power raising.
	 */
	bit = TOPFULL;

	while ((bit & power) == 0)
		bit >>= 1L;

	bit >>= 1L;

	res = ObjectCall(OBJ_SQUARE, vp, NULL_VALUE, NULL_VALUE);

	if (bit & power)
	{
		tmp = ObjectCall(OBJ_MUL, &res, vp, NULL_VALUE);
		objfree(res.v_obj);
		res = tmp;
	}

	bit >>= 1L;

	while (bit)
	{
		tmp = ObjectCall(OBJ_SQUARE, &res, NULL_VALUE, NULL_VALUE);

		objfree(res.v_obj);
		res = tmp;

		if (bit & power)
		{
			tmp = ObjectCall(OBJ_MUL, &res, vp, NULL_VALUE);
			objfree(res.v_obj);
			res = tmp;
		}

		bit >>= 1L;
	}

	if (qisneg(q))
	{
		tmp = ObjectCall(OBJ_INV, &res, NULL_VALUE, NULL_VALUE);
		objfree(res.v_obj);

		return tmp;
	}

	return res;
}


/*
 * Define a (possibly) new class of objects.
 * The list of indexes for the element names is also specified here,
 * and the number of elements defined for the object.  The definition
 * will fail if an attempt to redefine the object with different elements.
 * Returns TRUE if the object was successfully defined.
 *
 * given:
 *	name		name of object type
 *	indices		table of indices for elements
 *	count		number of elements defined for the object
 */
BOOL
DefineObject(const char * name, int indices[], int count)
{
	OBJECTACTIONS *		oap;
	STRINGHEAD *		hp;
	OBJECTACTIONS **	newobjects;
	int			index;

	hp = &objectNames;

	if (hp->h_list == NULL)
		initstr(hp);

	/*
	 * See if the object type is already defined.
	 */
	index = findstr(hp, name);

	if (index >= 0)
	{
		/*
		 * Object is already defined.  Give an error unless this
		 * new definition is exactly the same as the old one.
		 */
		oap = objects[index];

		if (oap->oa_count != count)
			return FALSE;

		for (index = 0; index < count; index++)
		{
			if (oap->oa_elements[index] != indices[index])
				return FALSE;
		}

		return TRUE;
	}

	/*
	 * This is a new definition.
	 * Allocate the object actions structure.
	 */
	if (hp->h_count >= maxObjectCount)
	{
		if (maxObjectCount == 0)
		{
			newobjects = (OBJECTACTIONS **) malloc(
				OBJALLOC * sizeof(OBJECTACTIONS *));

			maxObjectCount = OBJALLOC;
		}
		else
		{
			maxObjectCount += OBJALLOC;

			newobjects = (OBJECTACTIONS **) realloc(objects,
				maxObjectCount * sizeof(OBJECTACTIONS *));
		}

		if (newobjects == NULL)
		{
			math_error("Allocation failure for new object type");
			/*NOTREACHED*/
		}

		objects = newobjects;
	}

	oap = (OBJECTACTIONS *) malloc(objectactionsize(count));
	name = addstr(hp, name);

	if ((oap == NULL) || (name == NULL))
	{
		math_error("Cannot allocate object type");
		/*NOTREACHED*/
	}

	oap->oa_count = count;

	for (index = OBJ_MAXFUNC; index >= 0; index--)
		oap->oa_indices[index] = -1;

	for (index = 0; index < count; index++)
		oap->oa_elements[index] = indices[index];

	index = findstr(hp, name);
	oap->oa_index = index;
	objects[index] = oap;

	return TRUE;
}


/*
 * Check an object name to see if it is currently defined.
 * If so, the index for the object type is returned.
 * If the object name is currently unknown, then -1 is returned.
 */
int
checkobject(const char * name)
{
	STRINGHEAD *	hp;

	hp = &objectNames;

	if (hp->h_list == NULL)
		return -1;

	return findstr(hp, name);
}


/*
 * Define a (possibly) new element name for an object.
 * Returns an index which identifies the element name.
 */
int
addelement(const char * name)
{
	STRINGHEAD *	hp;
	int		index;

	hp = &elements;

	if (hp->h_list == NULL)
		initstr(hp);

	index = findstr(hp, name);

	if (index >= 0)
		return index;

	if (addstr(hp, name) == NULL)
	{
		math_error("Cannot allocate element name");
		/*NOTREACHED*/
	}

	return findstr(hp, name);
}


/*
 * Return the index which identifies an element name.
 * Returns minus one if the element name is unknown.
 *
 * given:
 *	name		element name
 */
int
findelement(const char * name)
{
	if (elements.h_list == NULL)
		return -1;

	return findstr(&elements, name);
}


/*
 * Returns the name of object type with specified index
 */
const char *
objtypename(unsigned long index)
{
	return namestr(&objectNames, (long)index);
}


/*
 * Return the value table offset to be used for an object element name.
 * This converts the element index from the element table into an offset
 * into the object value array.	 Returns -1 if the element index is unknown.
 */
int
objoffset(const OBJECT * op, long index)
{
	const OBJECTACTIONS *	oap;
	int			offset;		/* offset into value array */

	oap = op->o_actions;

	for (offset = oap->oa_count - 1; offset >= 0; offset--)
	{
		if (oap->oa_elements[offset] == index)
			return offset;
	}

	return -1;
}


/*
 * Allocate a new object structure with the specified index.
 */
OBJECT *
objalloc(long index)
{
	OBJECTACTIONS *	oap;
	OBJECT *	op;
	VALUE *		vp;
	int		i;

	if ((index < 0) || (index > maxObjectCount))
	{
		math_error("Allocating bad object index");
		/*NOTREACHED*/
	}

	oap = objects[index];

	if (oap == NULL)
	{
		math_error("Object type not defined");
		/*NOTREACHED*/
	}

	i = oap->oa_count;

	if (i < USUAL_ELEMENTS)
		i = USUAL_ELEMENTS;

	if (i == USUAL_ELEMENTS)
		op = (OBJECT *) malloc(sizeof(OBJECT));
	else
		op = (OBJECT *) malloc(objectsize(i));

	if (op == NULL)
	{
		math_error("Cannot allocate object");
		/*NOTREACHED*/
	}

	op->o_actions = oap;
	op->o_linkCount = 1;

	vp = op->o_table;

	for (i = oap->oa_count; i-- > 0; vp++)
	{
		vp->v_num = qlink(&_qzero_);
		vp->v_type = V_NUM;
		vp->v_subtype = V_NOSUBTYPE;
	}

	return op;
}


/*
 * Free an object structure.
 * This only really frees the object if its link count reaches zero.
 * This accepts a null pointer.
 */
void
objfree(OBJECT * op)
{
	VALUE *	vp;
	int	i;

	if (op == 0)
		return;

	if (op->o_linkCount <= 0)
	{
		math_error("Free of object with bad link count");
		/*NOTREACHED*/
	}

	if (--(op->o_linkCount) != 0)
		return;

	vp = op->o_table;

	for (i = op->o_actions->oa_count; i-- > 0; vp++)
	{
		if (vp->v_type == V_NUM)
		{
			qfree(vp->v_num);
		}
		else
		{
			freevalue(vp);
		}
	}

	free(op);
}


/*
 * Link to an object value.
 * After this call, another objfree call must be made before the object
 * is actually freed.  This accepts a null pointer.
 */
OBJECT *
ObjectLink(OBJECT * op)
{
	if (op != 0)
		op->o_linkCount++;

	return op;
}


/*
 * Copy an object value
 */
OBJECT *
objcopy(OBJECT * op)
{
	VALUE *		v1;
	VALUE *		v2;
	OBJECT *	np;
	int		i;

	i = op->o_actions->oa_count;

	if (i < USUAL_ELEMENTS)
		i = USUAL_ELEMENTS;

	if (i == USUAL_ELEMENTS)
		np = (OBJECT *) malloc(sizeof(OBJECT));
	else
		np = (OBJECT *) malloc(objectsize(i));

	if (np == NULL)
	{
		math_error("Cannot allocate object");
		/*NOTREACHED*/
	}

	np->o_actions = op->o_actions;
	np->o_linkCount = 1;

	v1 = op->o_table;
	v2 = np->o_table;

	for (i = op->o_actions->oa_count; i-- > 0; v1++, v2++)
	{
		if (v1->v_type == V_NUM)
		{
			v2->v_num = qlink(v1->v_num);
			v2->v_type = V_NUM;
		}
		else
		{
			copyvalue(v1, v2);
		}

		v2->v_subtype = V_NOSUBTYPE;
	}

	return np;
}


/*
 * Show object types that have been defined.
 */
void
showobjtypes(void)
{
	const STRINGHEAD *	hp;
	const OBJECTACTIONS *	oap;
	const STRINGHEAD *	ep;
	int			index;
	int			i;

	hp = &objectNames;
	ep = &elements;

	if (hp->h_count == 0)
	{
		printf("No object types defined\n");

		return;
	}

	for (index = 0; index < hp->h_count; index++)
	{
		oap = objects[index];

		printf("\t%s\t{", namestr(&objectNames, index));

		for (i = 0; i < oap->oa_count; i++)
		{
			if (i)
				printf(",");

			printf("%s", namestr(ep, oap->oa_elements[i]));
		}

		printf("}\n");
	}

}


/*
 * Method call to return the object's name.
 */
METHOD(m_ObjectType)
{
	VALUE   result;

	if (value->v_type != V_OBJ)
		math_error("Non-object value");

	result.v_str = makenewstring(
		objtypename(value->v_obj->o_actions->oa_index));

	result.v_type = V_STR;
	result.v_subtype = V_NOSUBTYPE;

	return result;
}

/* END CODE */
