/*
 * Node definition module.
 * This defines the opcode numbers, and those routines which implement
 * the opcodes.
 *
 * Copyright (C) 2009  David I. Bell
 *
 * Calc is open software; you can redistribute it and/or modify it under
 * the terms of the version 2.1 of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Calc is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU Lesser General
 * Public License for more details.
 *
 * A copy of version 2.1 of the GNU Lesser General Public License is
 * distributed with calc under the filename COPYING-LGPL.  You should have
 * received a copy with calc; if not, write to Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 */

#include "calc.h"
#include "node.h"
#include "userfunc.h"

/*
 * The maximum parsing depth required for a function.
 * This involves the depth of nested while, for, do, and switch statements
 * within the source code of the function being compiled, and has nothing
 * to do with the execution of the function later.
 */
#define	MAX_PARSE_DEPTH	200


/*
 * Flags set at parsing time which slightly affect the function both
 * during parsing and during execution.  Must functions have no flags set.
 */
#define	PARSE_FLAG_NONE		0x0000	/* no special handling */
#define	PARSE_FLAG_TOP_LEVEL	0x0001	/* this is the top-level function */
#define	PARSE_FLAG_AUTO_DEFINE	0x0002	/* automatically define symbols */
#define	PARSE_FLAG_AUTO_HISTORY	0x0004	/* automatically save in history */
#define	PARSE_FLAG_AUTO_RETURN	0x0008	/* automatically save return value */
#define	PARSE_FLAG_AUTO_PRINT	0x0010	/* automatically print expressions */
#define PARSE_FLAG_NO_DUMP	0x0020	/* don't dump nodes when defined */
#define	PARSE_FLAG_NO_MESSAGE	0x0040	/* don't print defined message */
#define	PARSE_FLAG_NO_ERRORS	0x0080	/* don't print compilation errors */
#define	PARSE_FLAG_PARAM_SYMS	0x0100	/* treat all symbols as parameters */
#define	PARSE_FLAG_EVALUATION	0x0200	/* only parse simple evaluations */
#define	PARSE_FLAG_ASSIGNMENT	0x0400	/* only parse single assignments */
#define	PARSE_FLAG_NO_OPTIMIZE	0x0800	/* don't optimize */


/*
 * Structure holding the current parse state for a user-defined function.
 */
typedef	struct
{
	const char *	funcName;	/* function name being parsed */
	Node *		nodeTree;	/* built node tree for function */
	int		parseFlags;	/* parsing flags */
	BOOL		isMethod;	/* parsing a method function */
	BOOL		hasFailed;	/* there was a parse failure */
	Node **		labelTable;	/* table of label nodes */
	int		labelCount;	/* number of label nodes in table */
	int		labelSize;	/* size of allocated label table */
	Node *		stack[MAX_PARSE_DEPTH];	/* parsing stack */
	int		stackDepth;	/* depth of parse stack */
} ParseState;


/*
 * The current parse state.
 */
extern	ParseState	parseState;


/*
 * Helper routines used while parsing and for node fixups after parsing.
 */
extern	void		SetParseState(const char *, BOOL, int);
extern	void		SetParseFailed(void);
extern	Node *		FindCaseOwner(void);
extern	Node *		FindBreakOwner(const char *);
extern	Node *		FindContinueOwner(const char *);
extern	Node *		FindLabeledStatement(const char *);
extern	Node *		FindLabelNode(const char *);
extern	int		FindLabelIndex(const char *);
extern	BOOL		AddLabelNode(Node *);
extern	BOOL		ValidateLabels(void);
extern	void		RecurseSwitchCaseJumpStates(Node *);
extern	void		RecurseBreakAndContinueAdjustments(Node *);
extern	BOOL		RecurseContainsLabelNode(const Node *);
extern	void		RecurseGotoAdjustments(Node *);
extern	BOOL		ContainsLabelNode(const Node *);
extern	UserFunc *	CompileExpression(int);
extern	UserFunc *	CompileImplicitFunction(int);
extern	UserFunc *	CompileNormalFunction(void);

/* END CODE */
