/*
 * Copyright (C) 2009  David I. Bell
 *
 * Calc is open software; you can redistribute it and/or modify it under
 * the terms of the version 2.1 of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Calc is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU Lesser General
 * Public License for more details.
 *
 * A copy of version 2.1 of the GNU Lesser General Public License is
 * distributed with calc under the filename COPYING-LGPL.  You should have
 * received a copy with calc; if not, write to Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Value history support for calc.
 *
 * This provides a list of history values for interactive use.
 * The number of values in the history is configurable so that the
 * user can trade off convenience for memory size.
 *
 * Note that the values in the history are copies of the calculated
 * values where possible.  However, for some types of values (such as
 * files), the copies are actually still links and so their internal
 * state can still change after the history value is stored.  Similarly,
 * when history values are referenced a copy is returned.
 *
 * There is always at least one value available for the history,
 * which is the '.' value, otherwise known as the "oldvalue".
 * This is the most recent value calculated by the command line.
 *
 * Besides the old value, the values calculated by the command line
 * are also stored in the value history.  Each calculated value is
 * assigned the next sequential integer id, starting from 1.  When
 * the maximum size of the history table is reached, then the earlier
 * history values are freed and are no longer available.
 *
 * Within an expression, the history values can be referenced using
 * the '$n' or '$-n' syntax in an expression where n is a non-negative
 * long value.  Positive values specify the particular history value
 * having the specified id.  Negative values are relative id values
 * used to reference history values back from the newest one, with -1
 * being the most recent history value.  The id value of zero is not used.
 *
 * Out of range or zero value ids return an indication of this fact so
 * that the callers can complain if required.  Otherwise a null value is
 * returned for such ids.
 */

#include "calc.h"
#include "value.h"


/*
 * The maximum allowed history table size.
 */
#define	MAX_VALUE_HISTORY_SIZE	100000


/*
 * The value history table information.
 */
static	long	historyBaseId = 1;	/* the base id for the history */
static	int	historyTableUsed;	/* number of values in table */
static	int	historyTableIndex;	/* base index of used entries */
static	int	historyTableSize;	/* allocated table size */
static	VALUE *	historyTable;		/* history table values */


/*
 * The "oldvalue" set by a top level calculation, which is always available.
 * There might be a linked value to this from the history table.
 */
static	VALUE	oldValue = {V_NULL, V_NOSUBTYPE};


/*
 * Clear the values in the history table and reset the history base id.
 * The size of the history table is not changed.
 */
void
ResetValueHistory(void)
{
	/*
	 * Free any old values.
	 */
	TrimValueHistory(0);

	/*
	 * Reset all values except for the table allocation.
	 */
	historyTableUsed = 0;
	historyTableIndex = 0;
	historyBaseId = 1;
}


/*
 * Reallocate the size of the history table so that it holds the
 * required number of elements.
 */
void
SetValueHistorySize(int size)
{
	VALUE *	oldTable;
	VALUE *	newTable;
	int	oldIndex;
	int	newIndex;
	int	count;

	/*
	 * Check for history size limits.
	 */
	if ((size < 0) || (size > MAX_VALUE_HISTORY_SIZE))
		math_error("Invalid value history size %d", size);

	/*
	 * If the size is not being changed then do nothing.
	 */
	if (size == historyTableSize)
		return;

	/*
	 * If the new size is zero then free it.
	 */
	if (size == 0)
	{
		TrimValueHistory(0);

		historyTableSize = 0;
		historyTableIndex = 0;
		historyTableUsed = 0;

		free(historyTable);
		historyTable = 0;

		return;
	}

	/*
	 * Trim the table down if the size requires it.
	 */
	TrimValueHistory(size);

	/*
	 * Allocate the new table.
	 */
	newTable = (VALUE *) malloc(size * sizeof(VALUE));

	if (newTable == 0)
		math_error("Cannot allocate value history table");

	/*
	 * Copy the values over from the old table.
	 */
	oldTable = historyTable;
	count = historyTableUsed;
	oldIndex = historyTableIndex;
	newIndex = 0;

	while (count-- > 0)
	{
		newTable[newIndex++] = oldTable[oldIndex++];

		if (oldIndex >= historyTableSize)
			oldIndex = 0;
	}

	/*
	 * Clear any new values in the new table.
	 */
	while (newIndex < size)
	{
		newTable[newIndex].v_type = V_NULL;
		newTable[newIndex].v_subtype = V_NOSUBTYPE;
		newIndex++;
	}

	/*
	 * Replace the old table.
	 */
	historyTableSize = 0;
	historyTableIndex = 0;
	historyTable = newTable;
	historyTableSize = size;

	/*
	 * Free the old table.
	 */
	free(oldTable);
}


/*
 * Trim the history table values down to the specified number of entries.
 * A size of zero will remove all of the history values.
 */
void
TrimValueHistory(int size)
{
	/*
	 * Fix illegal value.
	 */
	if (size < 0)
		size = 0;

	/*
	 * Free the history table entries one by one until its size
	 * reaches the goal.
	 */
	while (historyTableUsed > size)
	{
		/*
		 * Free the oldest entry.
		 */
		freevalue(historyTable + historyTableIndex);

		/*
		 * Reduce the count of entries left.
		 */
		historyTableUsed--;

		/*
		 * Advance the base index to the next entry.
		 */
		historyTableIndex = (historyTableIndex + 1) % historyTableSize;

		/*
		 * Increment the counter id for the base element.
		 */
		historyBaseId++;
	}
}


/*
 * Save the value as the old value and store it in the history if possible.
 * If the value is an error value then the last error value is also set.
 * The specified value is assumed to be the result of a calculation and
 * so is stored directly without copying.
 */
void
SaveOldValue(VALUE * value)
{
	int	newIndex;
	VALUE	lastValue;

	/*
	 * Store the value into the "old" value.
	 */
	freevalue(&oldValue);
	oldValue = *value;

	/*
	 * If this is an error then link it to the last error value.
	 */
	if (value->v_type == V_ERROR)
		SaveCurrentError(LinkError(value->v_error));

	/*
	 * If there is no history table then do nothing.
	 */
	if (historyTableSize == 0)
		return;

	/*
	 * If this value is the same as the latest value then do nothing.
	 */
	if (GetHistoryValue(-1L, &lastValue) &&
		!comparevalue(value, &lastValue))
	{
		return;
	}

	/*
	 * The new history value differs from the latest one.
	 * Trim the history if necessary to make room for one more value.
	 */
	TrimValueHistory(historyTableSize - 1);

	/*
	 * Get the index of the new entry.
	 */
	newIndex = (historyTableIndex + historyTableUsed) % historyTableSize;

	/*
	 * Link the value into the history entry.
	 */
	LinkValue(value, historyTable + newIndex);

	/*
	 * Increment the number of used entries.
	 */
	historyTableUsed++;
}


/*
 * Get the value of the "oldvalue".
 * The returned value needs to be copied if it is to be kept or modified.
 */
void
GetOldValue(VALUE * value)
{
	*value = oldValue;
}


/*
 * Get the value of one of the specified history values if possible.
 * The returned value needs to be copied if it is to be be kept or modified.
 * A positive value indicates an absolute history value.
 * A negative value indicates a relative history value from the last one.
 * Returns TRUE if the id value is legal and so a copy was made.
 * Otherwise a null value is returned and FALSE is returned.
 */
BOOL
GetHistoryValue(long id, VALUE * value)
{
	long	offset;
	int	index;

	/*
	 * Initialize the value to the null value in case of an error.
	 */
	value->v_type = V_NULL;
	value->v_subtype = V_NOSUBTYPE;

	/*
	 * The id value of zero is never valid.
	 */
	if (id == 0)
		return FALSE;

	/*
	 * If the id is relative then convert it to an absolute id.
	 */
	if (id < 0)
		id += historyBaseId + historyTableUsed;

	/*
	 * Calculate the offset from the base id.
	 */
	offset = id - historyBaseId;

	/*
	 * If the offset value is not within the number of entries in
	 * the table then fail.
	 */
	if ((id < 0) || (offset < 0) || (offset >= historyTableUsed))
		return FALSE;

	/*
	 * Get the index of the table entry at that offset.
	 */
	index = (historyTableIndex + ((int) offset)) % historyTableSize;

	/*
	 * Copy the selected value and return success.
	 */
	*value = historyTable[index];

	return TRUE;
}


/*
 * Return the current size of the history table.
 */
int
GetValueHistorySize(void)
{
	return historyTableSize;
}


/*
 * Return the number of entries in the history table.
 */
int
GetValueHistoryUsage(void)
{
	return historyTableUsed;
}


/*
 * Return the current base history id value.
 */
long
GetValueHistoryBaseId(void)
{
	return historyBaseId;
}


/*
 * Return the latest history id value.
 * Returns 0 if there is no value.
 */
long
GetValueHistoryLatestId(void)
{
	if (historyTableUsed == 0)
		return 0;

	return historyBaseId + historyTableUsed - 1;
}


/*
 * Show the values in the value history table.
 */
void
ShowValueHistory(void)
{
	int	count;
	int	index;
	VALUE *	value;
	char	buffer[24];

	/*
	 * If the table is empty then say so.
	 */
	if (historyTableUsed == 0)
	{
		math_str("\nNo history values\n\n");

		return;
	}

	/*
	 * Print all of the history values.
	 */
	math_chr('\n');

	for (count = 0; count < historyTableUsed; count++)
	{
		index = historyTableIndex + count;

		if (index >= historyTableSize)
			index -= historyTableSize;

		value = historyTable + index;

		sprintf(buffer, "$%ld:", historyBaseId + count);

		math_fmt("%6s ", buffer);

		fitprintvalue(value, 50);

		math_chr('\n');
	}

	math_chr('\n');
}
